import hashlib
from abc import ABC
from typing import Literal

from ._base import DataSelector


class ThreeDSelector(DataSelector, ABC):
    kind: Literal["3D"] = "3D"


class ThreeDModelFilteredSelector(ThreeDSelector):
    type: Literal["3DFiltered"] = "3DFiltered"
    model_type: Literal["Classic", "DataModel"] = "Classic"
    published: bool | None = None

    def __str__(self) -> str:
        suffix = f"3DModels_{self.model_type}"
        if self.published is not None:
            return f"{suffix}_published_{self.published}"
        return suffix


class ThreeDModelIdSelector(ThreeDSelector):
    type: Literal["3DId"] = "3DId"
    ids: tuple[int, ...]

    def __str__(self) -> str:
        hash_ = hashlib.md5(",".join(sorted(map(str, self.ids))).encode()).hexdigest()[:8]
        return f"3DModels_ids_count_{len(self.ids)}_hash_{hash_}"



import os
from pathlib import Path
from typing import ClassVar

from pydantic import BaseModel, Field, JsonValue

from ._base import YAMLFile
from ._types import ValidationType


class Environment(BaseModel):
    name: str = "dev"
    project: str = Field(default_factory=lambda: os.environ.get("CDF_PROJECT", "UNKNOWN"))
    validation_type: ValidationType = Field("dev", alias="validation-type")
    selected: list[str] | None = None


class ConfigYAML(YAMLFile):
    filename: ClassVar[str] = "config.{name}.yaml"

    environment: Environment = Field(default_factory=Environment)
    variables: dict[str, JsonValue] | None = None

    @classmethod
    def get_filename(cls, name: str) -> str:
        return cls.filename.format(name=name)

    @classmethod
    def get_filepath(cls, organization_dir: Path, name: str) -> Path:
        return organization_dir / cls.get_filename(name)



import builtins
from collections.abc import Iterable, Sequence
from typing import Any, Literal

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import (
    HTTPClient,
    ItemsSuccessResponse,
    RequestMessage,
    SuccessResponse,
)
from cognite_toolkit._cdf_tk.client.request_classes.filters import ClassicFilter
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling._references import InstanceId
from cognite_toolkit._cdf_tk.client.resource_classes.filemetadata import FileMetadataRequest, FileMetadataResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId, InternalId, InternalOrExternalId
from cognite_toolkit._cdf_tk.client.resource_classes.pending_instance_id import PendingInstanceId


class FileMetadataAPI(CDFResourceAPI[FileMetadataResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/files", item_limit=1, concurrency_max_workers=1),
                "retrieve": Endpoint(method="POST", path="/files/byids", item_limit=1000, concurrency_max_workers=1),
                "update": Endpoint(method="POST", path="/files/update", item_limit=1000, concurrency_max_workers=1),
                "delete": Endpoint(method="POST", path="/files/delete", item_limit=1000, concurrency_max_workers=1),
                "list": Endpoint(method="POST", path="/files/list", item_limit=1000),
            },
            api_version="alpha",
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[FileMetadataResponse]:
        return PagedResponse[FileMetadataResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[InternalOrExternalId]:
        return ResponseItems[InternalOrExternalId].model_validate_json(response.body)

    def create(self, items: Sequence[FileMetadataRequest], overwrite: bool = False) -> list[FileMetadataResponse]:
        """Upload file metadata to CDF.

        Args:
            items: List of FileMetadataRequest objects to upload.
            overwrite: Whether to overwrite existing file metadata with the same external ID.

        Returns:
            List of created FileMetadataResponse objects.
        """
        # The Files API is different from other APIs, thus we have a custom implementation here.
        # - It only allow one item per request that is not wrapped in a "items" field.
        # - It uses a query parameter for "overwrite" instead of including it in the body
        endpoint = self._method_endpoint_map["create"]
        results: list[FileMetadataResponse] = []
        for item in items:
            request = RequestMessage(
                endpoint_url=self._make_url(endpoint.path),
                method=endpoint.method,
                body_content=item.dump(),
                parameters={"overwrite": overwrite},
            )
            response = self._http_client.request_single_retries(request)
            result = response.get_success_or_raise()
            results.append(FileMetadataResponse.model_validate_json(result.body))
        return results

    def retrieve(
        self, items: Sequence[InternalId | ExternalId | InstanceId], ignore_unknown_ids: bool = False
    ) -> list[FileMetadataResponse]:
        """Retrieve file metadata from CDF.

        Args:
            items: List of InternalOrExternalId objects to retrieve.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        Returns:
            List of retrieved FileMetadataResponse objects.
        """
        return self._request_item_response(
            items, method="retrieve", extra_body={"ignoreUnknownIds": ignore_unknown_ids}
        )

    def update(
        self, items: Sequence[FileMetadataRequest], mode: Literal["patch", "replace"] = "replace"
    ) -> list[FileMetadataResponse]:
        """Update file metadata in CDF.

        Args:
            items: List of FileMetadataRequest objects to update.
            mode: Update mode, either "patch" or "replace".

        Returns:
            List of updated FileMetadataResponse objects.
        """
        return self._update(items, mode=mode)

    def delete(self, items: Sequence[InternalOrExternalId], ignore_unknown_ids: bool = False) -> None:
        """Delete file metadata from CDF.

        Args:
            items: List of InternalOrExternalId objects to delete.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        """
        self._request_no_response(items, "delete", extra_body={"ignoreUnknownIds": ignore_unknown_ids})

    def paginate(
        self,
        filter: ClassicFilter | None = None,
        directory_prefix: str | None = None,
        uploaded: bool | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[FileMetadataResponse]:
        """Iterate over file metadata in CDF.

        Args:
            filter: Filter by data set IDs and/or asset subtree IDs.
            directory_prefix: Filter by directory prefix.
            uploaded: Filter by upload status.
            limit: Maximum number of items to return per page.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of FileMetadataResponse objects.
        """
        filter_: dict[str, Any] = filter.dump() if filter else {}
        if directory_prefix is not None:
            filter_["directoryPrefix"] = directory_prefix
        if uploaded is not None:
            filter_["uploaded"] = uploaded

        return self._paginate(
            cursor=cursor,
            limit=limit,
            body={"filter": filter_ or None},
        )

    def iterate(
        self,
        filter: ClassicFilter | None = None,
        directory_prefix: str | None = None,
        uploaded: bool | None = None,
        limit: int | None = 100,
    ) -> Iterable[list[FileMetadataResponse]]:
        """Iterate over file metadata in CDF.

        Args:
            filter: Filter by data set IDs and/or asset subtree IDs.
            directory_prefix: Filter by directory prefix.
            uploaded: Filter by upload status.
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of FileMetadataResponse objects.
        """
        filter_: dict[str, Any] = filter.dump() if filter else {}
        if directory_prefix is not None:
            filter_["directoryPrefix"] = directory_prefix
        if uploaded is not None:
            filter_["uploaded"] = uploaded

        return self._iterate(
            limit=limit,
            body={"filter": filter_ or None},
        )

    def list(
        self,
        limit: int | None = 100,
    ) -> list[FileMetadataResponse]:
        """List all file metadata in CDF.

        Args:
            limit: Maximum number of items to return. None for all items.

        Returns:
            List of FileMetadataResponse objects.
        """
        return self._list(limit=limit)

    def set_pending_ids(self, items: Sequence[PendingInstanceId]) -> builtins.list[FileMetadataResponse]:
        """Set pending instance IDs for one or more file metadata entries.

        This links asset-centric files to DM nodes that will be created
        by the syncer service.

        Args:
            items: Sequence of PendingInstanceId objects containing the pending
                instance IDs and the file id or external_id to link them to.

        Returns:
            List of updated FileMetadataResponse objects.
        """
        return self._request_item_response(items, method="retrieve", endpoint="/files/set-pending-instance-ids")

    def unlink_instance_ids(self, items: Sequence[InternalOrExternalId]) -> builtins.list[FileMetadataResponse]:
        """Unlink instance IDs from files.

        This allows a CogniteFile node in Data Modeling to be deleted
        without deleting the underlying file content.

        Args:
            items: Sequence of InternalOrExternalId identifying the files to unlink.

        Returns:
            List of updated FileMetadataResponse objects.
        """
        return self._request_item_response(items, method="retrieve", endpoint="/files/unlink-instance-ids")



from ._download import DownloadCommand
from ._migrate import (
    MigrationCommand,
    MigrationPrepareCommand,
)
from ._profile import ProfileAssetCentricCommand, ProfileAssetCommand, ProfileRawCommand, ProfileTransformationCommand
from ._purge import PurgeCommand
from ._respace import RespaceCommand
from ._upload import UploadCommand
from .about import AboutCommand
from .auth import AuthCommand
from .build_cmd import BuildCommand
from .build_v2.build_v2 import BuildV2Command
from .clean import CleanCommand
from .collect import CollectCommand
from .deploy import DeployCommand
from .dump_resource import DumpResourceCommand
from .init import InitCommand
from .modules import ModulesCommand
from .pull import PullCommand
from .repo import RepoCommand
from .resources import ResourcesCommand
from .run import RunFunctionCommand, RunTransformationCommand, RunWorkflowCommand

__all__ = [
    "AboutCommand",
    "AuthCommand",
    "BuildCommand",
    "BuildV2Command",
    "CleanCommand",
    "CollectCommand",
    "DeployCommand",
    "DownloadCommand",
    "DumpResourceCommand",
    "InitCommand",
    "MigrationCommand",
    "MigrationPrepareCommand",
    "ModulesCommand",
    "ProfileAssetCentricCommand",
    "ProfileAssetCommand",
    "ProfileRawCommand",
    "ProfileTransformationCommand",
    "PullCommand",
    "PurgeCommand",
    "RepoCommand",
    "ResourcesCommand",
    "RespaceCommand",
    "RunFunctionCommand",
    "RunTransformationCommand",
    "RunWorkflowCommand",
    "UploadCommand",
]



from abc import ABC
from typing import Any

from pydantic import field_serializer
from pydantic_core.core_schema import FieldSerializationInfo

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject, RequestResource, ResponseResource

from ._references import DataModelReference, ViewReference
from ._view import ViewResponse


class DataModel(BaseModelObject, ABC):
    """Cognite Data Model resource.

    Data models group and structure views into reusable collections.
    A data model contains a set of views where the node types can
    refer to each other with direct relations and edges.
    """

    space: str
    external_id: str
    version: str
    name: str | None = None
    description: str | None = None

    def as_id(self) -> DataModelReference:
        return DataModelReference(
            space=self.space,
            external_id=self.external_id,
            version=self.version,
        )


class DataModelRequest(DataModel, RequestResource):
    views: list[ViewReference] | None = None

    @field_serializer("views", mode="plain")
    @classmethod
    def serialize_views(
        cls, views: list[ViewReference] | None, info: FieldSerializationInfo
    ) -> list[dict[str, Any]] | None:
        if views is None:
            return None
        return [{**view.model_dump(**vars(info)), "type": "view"} for view in views]


class DataModelResponse(DataModel, ResponseResource[DataModelRequest]):
    views: list[ViewReference] | None = None
    created_time: int
    last_updated_time: int
    is_global: bool

    @classmethod
    def request_cls(cls) -> type[DataModelRequest]:
        return DataModelRequest

    @field_serializer("views", mode="plain")
    @classmethod
    def serialize_views(
        cls, views: list[ViewReference] | None, info: FieldSerializationInfo
    ) -> list[dict[str, Any]] | None:
        if views is None:
            return None
        return [{**view.model_dump(**vars(info)), "type": "view"} for view in views]


class DataModelResponseWithViews(DataModel, ResponseResource[DataModelRequest]):
    views: list[ViewResponse] | None = None
    created_time: int
    last_updated_time: int
    is_global: bool

    @classmethod
    def request_cls(cls) -> type[DataModelRequest]:
        return DataModelRequest



import sys
from collections import UserList

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from pydantic import BaseModel, Field

from cognite_toolkit._cdf_tk.tk_warnings import ToolkitWarning, WarningList

MODULE_ISSUE_CODE = "MOD"


class Issue(BaseModel):
    """Base class for all issues"""

    code: str
    message: str | None = Field(default=None)


# temporary adapter to manage existing warnings
class IssueList(UserList[Issue]):
    """List of build issues."""

    @classmethod
    def from_warning_list(cls, warning_list: WarningList[ToolkitWarning]) -> Self:
        """Create a IssueList from a WarningList."""
        return cls([Issue(code="WARN", message=warning.get_message()) for warning in warning_list])


class ModuleLoadingIssue(Issue):
    """Issue with the loading of a module folder."""

    code: str = "MOD_001"



from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject, RequestResource, ResponseResource

from .data_modeling import ViewReferenceNoVersion


class SearchConfigViewProperty(BaseModelObject):
    """Configuration for a property in search configuration."""

    property: str
    disabled: bool | None = None
    selected: bool | None = None
    hidden: bool | None = None


class SearchConfigBase(BaseModelObject):
    """Base class for search configuration with common fields."""

    view: ViewReferenceNoVersion
    use_as_name: str | None = None
    use_as_description: str | None = None
    columns_layout: list[SearchConfigViewProperty] | None = None
    filter_layout: list[SearchConfigViewProperty] | None = None
    properties_layout: list[SearchConfigViewProperty] | None = None

    def as_id(self) -> ViewReferenceNoVersion:
        return self.view


class SearchConfigRequest(SearchConfigBase, RequestResource):
    """Request resource for creating/updating search configuration."""

    # This is required when updating an existing search config
    id: int | None = None


class SearchConfigResponse(SearchConfigBase, ResponseResource[SearchConfigRequest]):
    """Response resource for search configuration."""

    id: int
    created_time: int
    last_updated_time: int

    @classmethod
    def request_cls(cls) -> type[SearchConfigRequest]:
        return SearchConfigRequest



from collections.abc import Iterable, Sequence

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.request_classes.filters import SimulatorModelRoutineFilter
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import InternalOrExternalId
from cognite_toolkit._cdf_tk.client.resource_classes.simulator_routine import (
    SimulatorRoutineRequest,
    SimulatorRoutineResponse,
)


class SimulatorRoutinesAPI(CDFResourceAPI[SimulatorRoutineResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/simulators/routines", item_limit=1),
                "delete": Endpoint(method="POST", path="/simulators/routines/delete", item_limit=1),
                "list": Endpoint(method="POST", path="/simulators/routines/list", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[SimulatorRoutineResponse]:
        return PagedResponse[SimulatorRoutineResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[InternalOrExternalId]:
        return ResponseItems[InternalOrExternalId].model_validate_json(response.body)

    def create(self, items: Sequence[SimulatorRoutineRequest]) -> list[SimulatorRoutineResponse]:
        """Create simulator routines in CDF.

        Args:
            items: List of SimulatorRoutineRequest objects to create.

        Returns:
            List of created SimulatorRoutineResponse objects.
        """
        return self._request_item_response(items, "create")

    def delete(self, items: Sequence[InternalOrExternalId]) -> None:
        """Delete simulator routines from CDF.

        Args:
            items: List of InternalOrExternalId objects to delete.
        """
        self._request_no_response(items, "delete")

    def paginate(
        self,
        filter: SimulatorModelRoutineFilter | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[SimulatorRoutineResponse]:
        """Iterate over simulator routines in CDF.

        Args:
            filter: Filter to apply to the simulator routines.
            limit: Maximum number of items to return per page.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of SimulatorRoutineResponse objects.
        """
        return self._paginate(
            body={"filter": filter.dump() if filter else None},
            cursor=cursor,
            limit=limit,
        )

    def iterate(
        self,
        filter: SimulatorModelRoutineFilter | None = None,
        limit: int | None = 100,
    ) -> Iterable[list[SimulatorRoutineResponse]]:
        """Iterate over simulator routines in CDF.

        Args:
            filter: Filter to apply to the simulator routines.
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of SimulatorRoutineResponse objects.
        """
        return self._iterate(
            limit=limit,
            body={
                "filter": filter.dump() if filter else None,
            },
        )

    def list(
        self,
        limit: int | None = 100,
    ) -> list[SimulatorRoutineResponse]:
        """List all simulator routines in CDF.

        Args:
            limit: Maximum number of items to return. None for all items.

        Returns:
            List of SimulatorRoutineResponse objects.
        """
        return self._list(limit=limit)



import time
from collections import defaultdict
from collections.abc import Hashable, Iterable, Sequence
from functools import cached_property
from pathlib import Path
from typing import Any, Literal, cast, final

from cognite.client.data_classes import ClientCredentials
from cognite.client.data_classes.capabilities import (
    AllScope,
    Capability,
    DataSetScope,
    FilesAcl,
    FunctionsAcl,
    SessionsAcl,
)
from cognite.client.data_classes.data_modeling import NodeId
from cognite.client.data_classes.data_modeling.cdm.v1 import CogniteFileApply
from cognite.client.data_classes.functions import HANDLER_FILE_NAME
from cognite.client.exceptions import CogniteAPIError
from rich import print
from rich.console import Console

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.resource_classes.function import FunctionRequest, FunctionResponse
from cognite_toolkit._cdf_tk.client.resource_classes.function_schedule import (
    FunctionScheduleId,
    FunctionScheduleRequest,
    FunctionScheduleResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId, InternalId
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.exceptions import (
    ResourceCreationError,
    ToolkitRequiredValueError,
)
from cognite_toolkit._cdf_tk.resource_classes import FunctionScheduleYAML, FunctionsYAML
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning, LowSeverityWarning
from cognite_toolkit._cdf_tk.utils import (
    calculate_directory_hash,
    calculate_hash,
    calculate_secure_hash,
    humanize_collection,
)
from cognite_toolkit._cdf_tk.utils.cdf import read_auth, try_find_error
from cognite_toolkit._cdf_tk.utils.file import create_temporary_zip, sanitize_filename
from cognite_toolkit._cdf_tk.utils.text import suffix_description

from .auth import GroupAllScopedCRUD
from .data_organization import DataSetsCRUD
from .group_scoped import GroupResourceScopedCRUD


@final
class FunctionCRUD(ResourceCRUD[ExternalId, FunctionRequest, FunctionResponse]):
    support_drop = True
    folder_name = "functions"
    resource_cls = FunctionResponse
    resource_write_cls = FunctionRequest
    kind = "Function"
    yaml_cls = FunctionsYAML
    dependencies = frozenset({DataSetsCRUD, GroupAllScopedCRUD})
    _doc_url = "Functions/operation/postFunctions"
    metadata_value_limit = 512
    support_update = False

    class _MetadataKey:
        function_hash = "cognite-toolkit-hash"
        secret_hash = "cdf-toolkit-secret-hash"

    def __init__(self, client: ToolkitClient, build_path: Path | None, console: Console | None):
        super().__init__(client, build_path, console)
        self.data_set_id_by_external_id: dict[str, int] = {}
        self.space_by_external_id: dict[str, str] = {}
        self.function_dir_by_external_id: dict[str, Path] = {}

    @cached_property
    def project_data_modeling_status(self) -> Literal["HYBRID", "DATA_MODELING_ONLY"]:
        status = self.client.project.status()
        return status.this_project.data_modeling_status

    @property
    def display_name(self) -> str:
        return "functions"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[FunctionRequest] | None, read_only: bool
    ) -> list[Capability] | list[Capability]:
        if not items and items is not None:
            return []

        function_actions = (
            [FunctionsAcl.Action.Read] if read_only else [FunctionsAcl.Action.Read, FunctionsAcl.Action.Write]
        )
        file_actions = [FilesAcl.Action.Read] if read_only else [FilesAcl.Action.Read, FilesAcl.Action.Write]

        return [
            FunctionsAcl(function_actions, FunctionsAcl.Scope.All()),
            FilesAcl(file_actions, FilesAcl.Scope.All()),  # Needed for uploading function artifacts
        ]

    @classmethod
    def get_id(cls, item: FunctionResponse | FunctionRequest | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        if item.external_id is None:
            raise ToolkitRequiredValueError("Function must have external_id set.")
        return ExternalId(external_id=item.external_id)

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def as_str(cls, id: ExternalId) -> str:
        return sanitize_filename(id.external_id)

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "dataSetExternalId" in item:
            yield DataSetsCRUD, ExternalId(external_id=item["dataSetExternalId"])

    def load_resource_file(
        self, filepath: Path, environment_variables: dict[str, str | None] | None = None
    ) -> list[dict[str, Any]]:
        if filepath.parent.name != self.folder_name:
            # Functions configs needs to be in the root function folder.
            # This is to allow arbitrary YAML files inside the function code folder.
            return []
        if self.resource_build_path is None:
            raise ValueError("build_path must be set to compare functions as function code must be compared.")

        raw_list = super().load_resource_file(filepath, environment_variables)
        for item in raw_list:
            item_id = self.get_id(item)
            external_id = item_id.external_id
            function_rootdir = Path(self.resource_build_path / external_id)
            self.function_dir_by_external_id[external_id] = function_rootdir
            if "metadata" not in item:
                item["metadata"] = {}
            value = self._create_hash_values(function_rootdir)
            item["metadata"][self._MetadataKey.function_hash] = value
            if "secrets" in item:
                item["metadata"][self._MetadataKey.secret_hash] = calculate_secure_hash(item["secrets"])

        return raw_list

    @classmethod
    def _create_hash_values(cls, function_rootdir: Path) -> str:
        root_hash = calculate_directory_hash(
            function_rootdir, exclude_prefixes={".DS_Store"}, ignore_files={".pyc"}, shorten=True
        )
        hash_value = f"/={root_hash}"
        to_search = [function_rootdir]
        while to_search:
            search_dir = to_search.pop()
            for file in sorted(search_dir.glob("*"), key=lambda x: x.relative_to(function_rootdir).as_posix()):
                if file.is_dir():
                    to_search.append(file)
                    continue
                elif file.is_file() and file.suffix == ".pyc":
                    continue
                elif file.is_file() and file.name == ".DS_Store":
                    continue
                file_hash = calculate_hash(file, shorten=True)
                new_entry = f"{file.relative_to(function_rootdir).as_posix()}={file_hash}"
                if len(hash_value) + len(new_entry) > (cls.metadata_value_limit - 1):
                    break
                hash_value += f";{new_entry}"
        return hash_value

    def get_function_required_capabilities(
        self, items: Sequence[FunctionRequest] | None, read_only: bool
    ) -> list[Capability]:
        """
        Get required capabilities for working with CDF Functions and their associated files.

        This method determines the necessary permissions needed for function operations,
        with dataset-scoped file access when possible. When functions are associated with
        specific datasets, it will request file permissions only for those datasets
        rather than requiring all-scope file access.
        """
        if items is not None and not items:
            return []

        function_actions = (
            [FunctionsAcl.Action.Read] if read_only else [FunctionsAcl.Action.Read, FunctionsAcl.Action.Write]
        )
        file_actions = [FilesAcl.Action.Read] if read_only else [FilesAcl.Action.Read, FilesAcl.Action.Write]

        file_scope: AllScope | DataSetScope = FilesAcl.Scope.All()

        if items and self.data_set_id_by_external_id:
            dataset_ids = [
                self.data_set_id_by_external_id[item.external_id]
                for item in items
                if item.external_id and item.external_id in self.data_set_id_by_external_id
            ]
            if dataset_ids:
                file_scope = FilesAcl.Scope.DataSet(dataset_ids)

        return [
            FunctionsAcl(function_actions, FunctionsAcl.Scope.All()),
            FilesAcl(file_actions, file_scope),  # Needed for uploading function artifacts
        ]

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> FunctionRequest:
        item_id = self.get_id(resource)
        external_id = item_id.external_id
        if ds_external_id := resource.pop("dataSetExternalId", None):
            self.data_set_id_by_external_id[external_id] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        if space := resource.pop("space", None):
            self.space_by_external_id[external_id] = space
        # The fileID is required for the function to be created, but in the `.create` method
        # we first create that file and then set the fileID.
        if "fileId" not in resource or not isinstance(resource.get("fileId"), int):
            resource["fileId"] = -1  # Placeholder, will be set in create()
        return FunctionRequest.model_validate(resource)

    def dump_resource(self, resource: FunctionResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        if resource.status == "Failed":
            dumped = self.dump_id(ExternalId(external_id=resource.external_id or resource.name))
            dumped["status"] = "Failed"
            return dumped
        dumped = resource.as_request_resource().dump()
        local = local or {}
        for key in ["cpu", "memory", "runtime"]:
            if key not in local:
                # Server set default values
                dumped.pop(key, None)
            elif isinstance(local.get(key), float) and local[key] < dumped.get(key, 0):
                # On Azure and AWS, the server sets the CPU and Memory to the default values if the user
                # pass in lower values. We set this to match the local to avoid triggering a redeploy.
                # Note the user will get a warning about this when the function is created.
                if self.client.config.cloud_provider in ("azure", "aws"):
                    dumped[key] = local[key]
                elif self.client.config.cloud_provider == "gcp" and key == "cpu" and local[key] < 1.0:
                    # GCP does not allow CPU to be set to below 1.0
                    dumped[key] = local[key]
                elif self.client.config.cloud_provider == "gcp" and key == "memory" and local[key] < 1.5:
                    # GCP does not allow Memory to be set to below 1.5
                    dumped[key] = local[key]

        for key in ["indexUrl", "extraIndexUrls"]:
            # Only in write (request) format of the function
            if key in local:
                dumped[key] = local[key]
        if "secrets" in dumped and "secrets" not in local:
            # Secrets are masked in the response.
            dumped.pop("secrets")
        elif "secrets" in dumped and "secrets" in local:
            # Note this will be misleading as the secrets might not be the same as the ones in the API.
            # It will be caught by the hash comparison, but can cause confusion for the user if they
            # compare the dumped file with the API.
            dumped["secrets"] = local["secrets"]

        if file_id := dumped.pop("fileId", None):
            # The fileId is not part of the write object.
            function_zip_file = self.client.files.retrieve(id=file_id)
            if function_zip_file and (data_set_id := function_zip_file.data_set_id):
                dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
            if function_zip_file and function_zip_file.instance_id is not None:
                dumped["space"] = function_zip_file.instance_id.space

        if dumped.get("functionPath") == HANDLER_FILE_NAME and "functionPath" not in local:
            # Remove the default value of the functionPath
            dumped.pop("functionPath", None)

        # The Function API removes none-like values when processing the request.
        # Ref https://github.com/cognitedata/infrastructure/blob/7459a8f6959af89240c3250f9133c079a19d57c5/services/context/context_api/utils/schema_helpers.py#L8
        # This ensures that if the user sets owner or description to None or empty string,
        # it will not cause the comparison with the server side to evaluate as different,
        # even though the function is unchanged.
        for key in ["owner", "description"]:
            if key in local and local[key] in {None, ""} and key not in dumped:
                dumped[key] = local[key]

        if dumped.get("runtime") == "py312" and "runtime" not in local:
            # Remove the default value of the runtime for Python 3.12, as it is the default runtime and does not need to be specified.
            dumped.pop("runtime", None)

        return dumped

    def _is_activated(self, action: str) -> bool:
        status = self.client.functions.status()
        if status.status == "activated":
            return True
        if status.status == "requested":
            print(
                HighSeverityWarning(
                    f"Function service activation is in progress, cannot {action} functions."
                ).get_message()
            )
            return False
        else:
            print(
                HighSeverityWarning(
                    "Function service has not been activated, activating now, this may take up to 2 hours..."
                ).get_message()
            )
            self.client.functions.activate()
        return False

    def create(self, items: Sequence[FunctionRequest]) -> list[FunctionResponse]:
        created: list[FunctionResponse] = []
        if not self._is_activated("create"):
            return created
        if self.resource_build_path is None:
            raise ValueError("build_path must be set to compare functions as function code must be compared.")
        for item in items:
            external_id = item.external_id or item.name
            file_id = self._upload_function_code(external_id, item)
            # Wait until the files is available
            t0 = time.perf_counter()
            sleep_time = 1.0  # seconds
            for i in range(6):
                file = self.client.files.retrieve(id=file_id)
                if file and file.uploaded:
                    break
                time.sleep(sleep_time)
                sleep_time *= 2
            else:
                elapsed_time = time.perf_counter() - t0
                raise ResourceCreationError(
                    f"Failed to create function {external_id}. CDF API timed out after {elapsed_time:.0f} "
                    "seconds while waiting for the function code to be uploaded. Wait and try again.\nIf the"
                    " problem persists, please contact Cognite support."
                )
            # Create a copy with the file_id set
            item_to_create = FunctionRequest.model_validate({**item.dump(), "fileId": file_id})
            result = self.client.tool.functions.create([item_to_create])
            if result:
                created_item = result[0]
                self._warn_if_cpu_or_memory_changed(created_item, item)
                created.append(created_item)
        return created

    def _upload_function_code(self, external_id: str, item: FunctionRequest) -> int:
        """Uploads the function code to CDF.

        It will either upload the code to a CogniteFile if a space is provided and the feature flag is enabled,
        or to a classic file if not.

        Args:
            external_id (str): The external ID of the function.
            item (FunctionWrite): The function write object.

        Returns:
            int: The ID of the created file metadata object.
        """
        function_rootdir = self.function_dir_by_external_id[external_id]
        data_set_id = self.data_set_id_by_external_id.get(external_id)
        space = self.space_by_external_id.get(external_id)
        if space is None and self.project_data_modeling_status.casefold() == "data_modeling_only":
            raise ResourceCreationError(
                f"Function {external_id!r} must have a space set when the project is in DATA_MODELING_ONLY mode. "
                "This is used to set the NodeId of the CogniteFile created for the function code."
            )

        with create_temporary_zip(function_rootdir, "function.zip") as zip_path:
            if space:
                cognite_file = CogniteFileApply(
                    space=space,
                    external_id=external_id,
                    name=f"{sanitize_filename(item.name)}.zip",
                    mime_type="application/zip",
                )
                _ = self.client.data_modeling.instances.apply(cognite_file, replace=True)
                upload_file = self.client.files.upload_content_bytes(
                    zip_path.read_bytes(), instance_id=cognite_file.as_id()
                )
            else:
                upload_file = self.client.files.upload_bytes(
                    zip_path.read_bytes(),
                    name=f"{sanitize_filename(item.name)}.zip",
                    external_id=external_id,
                    overwrite=True,
                    data_set_id=data_set_id,
                )
        return upload_file.id

    @staticmethod
    def _warn_if_cpu_or_memory_changed(created_item: FunctionResponse, item: FunctionRequest) -> None:
        is_cpu_increased = (
            isinstance(item.cpu, float) and isinstance(created_item.cpu, float) and item.cpu < created_item.cpu
        )
        is_mem_increased = (
            isinstance(item.memory, float)
            and isinstance(created_item.memory, float)
            and item.memory < created_item.memory
        )
        if is_cpu_increased and is_mem_increased:
            prefix = "CPU and Memory"
            suffix = f"CPU {item.cpu} -> {created_item.cpu}, Memory {item.memory} -> {created_item.memory}"
        elif is_cpu_increased:
            prefix = "CPU"
            suffix = f"{item.cpu} -> {created_item.cpu}"
        elif is_mem_increased:
            prefix = "Memory"
            suffix = f"{item.memory} -> {created_item.memory}"
        else:
            return
        # The server sets the CPU and Memory to the default values, if the user pass in a lower value.
        # This happens on Azure and AWS. Warning the user about this.
        LowSeverityWarning(
            f"Function {prefix} is not configurable. Function {item.external_id!r} set {suffix}"
        ).print_warning()

    def retrieve(self, ids: Sequence[ExternalId]) -> list[FunctionResponse]:
        if not self._is_activated("retrieve"):
            return []
        return self.client.tool.functions.retrieve(list(ids), ignore_unknown_ids=True)

    def delete(self, ids: Sequence[ExternalId]) -> int:
        functions = self.retrieve(ids)

        self.client.tool.functions.delete(list(ids), ignore_unknown_ids=True)
        file_ids = {func.file_id for func in functions if func.file_id}
        files = self.client.files.retrieve_multiple(list(file_ids), ignore_unknown_ids=True)
        dm_file_nodes: set[NodeId] = set()
        classic_file_ids: set[int] = set()
        for file in files:
            if file.instance_id is not None:
                dm_file_nodes.add(file.instance_id)
            else:
                classic_file_ids.add(file.id)
        if classic_file_ids:
            self.client.files.delete(id=list(classic_file_ids), ignore_unknown_ids=True)
        if dm_file_nodes:
            self.client.data_modeling.instances.delete(list(dm_file_nodes))
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[FunctionResponse]:
        for functions in self.client.tool.functions.iterate():
            yield from functions


@final
class FunctionScheduleCRUD(ResourceCRUD[FunctionScheduleId, FunctionScheduleRequest, FunctionScheduleResponse]):
    folder_name = "functions"
    resource_cls = FunctionScheduleResponse
    resource_write_cls = FunctionScheduleRequest
    kind = "Schedule"
    yaml_cls = FunctionScheduleYAML
    dependencies = frozenset({FunctionCRUD, GroupResourceScopedCRUD, GroupAllScopedCRUD})
    _doc_url = "Function-schedules/operation/postFunctionSchedules"
    parent_resource = frozenset({FunctionCRUD})
    support_update = False

    _hash_key = "cdf-auth"
    _description_character_limit = 500

    def __init__(self, client: ToolkitClient, build_path: Path | None, console: Console | None):
        super().__init__(client, build_path, console)
        self.authentication_by_id: dict[FunctionScheduleId, ClientCredentials] = {}

    @property
    def display_name(self) -> str:
        return "function schedules"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[FunctionScheduleRequest] | None, read_only: bool
    ) -> list[Capability]:
        if not items and items is not None:
            return []

        function_actions = (
            [FunctionsAcl.Action.Read] if read_only else [FunctionsAcl.Action.Read, FunctionsAcl.Action.Write]
        )
        required_capabilities: list[Capability] = [
            FunctionsAcl(function_actions, FunctionsAcl.Scope.All()),
        ]
        if not read_only:
            required_capabilities.append(SessionsAcl([SessionsAcl.Action.Create], SessionsAcl.Scope.All()))
        return required_capabilities

    @classmethod
    def dump_id(cls, id: FunctionScheduleId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_id(cls, item: FunctionScheduleRequest | FunctionScheduleResponse | dict) -> FunctionScheduleId:
        if isinstance(item, dict):
            if missing := tuple(k for k in {"functionExternalId", "name"} if k not in item):
                # We need to raise a KeyError with all missing keys to get the correct error message.
                raise KeyError(*missing)
            return FunctionScheduleId(function_external_id=item["functionExternalId"], name=item["name"])

        if item.function_external_id is None or item.name is None:
            raise ToolkitRequiredValueError("FunctionSchedule must have functionExternalId and Name set.")
        return FunctionScheduleId(function_external_id=item.function_external_id, name=item.name)

    @classmethod
    def as_str(cls, id: FunctionScheduleId) -> str:
        return sanitize_filename(f"{id.function_external_id}-{id.name}")

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "functionExternalId" in item:
            yield FunctionCRUD, ExternalId(external_id=item["functionExternalId"])

    def load_resource_file(
        self, filepath: Path, environment_variables: dict[str, str | None] | None = None
    ) -> list[dict[str, Any]]:
        resources = super().load_resource_file(filepath, environment_variables)
        # We need to the auth hash calculation here, as the output of the load_resource_file
        # is used to compare with the CDF resource.
        for resource in resources:
            identifier = self.get_id(resource)
            credentials = read_auth(
                resource.get("authentication"),
                self.client.config,
                identifier,
                "function schedule",
                console=self.console,
            )
            self.authentication_by_id[identifier] = credentials
            auth_hash = calculate_secure_hash(credentials.dump(camel_case=True), shorten=True)
            extra_str = f"{self._hash_key}: {auth_hash}"
            resource["description"] = suffix_description(
                extra_str,
                resource.get("description"),
                self._description_character_limit,
                self.get_id(resource),
                self.display_name,
                self.console,
            )
        return resources

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> FunctionScheduleRequest:
        if "functionId" in resource:
            identifier = self.get_id(resource)
            LowSeverityWarning(f"FunctionId will be ignored in the schedule {identifier!r}").print_warning(
                console=self.console
            )
            resource.pop("functionId", None)
        # Remove authentication from resource dict as it contains clientId/clientSecret
        # which are stored separately in _authentication_by_id and converted to nonce at creation time
        resource.pop("authentication", None)
        return FunctionScheduleRequest.model_validate(resource)

    def dump_resource(self, resource: FunctionScheduleResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        # Dump the response fields directly since as_request_resource() is not supported
        dumped = resource.as_request_resource().dump()
        # Add functionExternalId to the dumped resource for comparison and retrieval purposes, even though it is not part of the API response.
        dumped["functionExternalId"] = resource.function_external_id
        local = local or {}
        if "functionId" in dumped and "functionId" not in local:
            dumped.pop("functionId")
        if "authentication" in local:
            # The authentication is not returned in the response, so we need to add it back.
            dumped["authentication"] = local["authentication"]
        return dumped

    def retrieve(self, ids: Sequence[FunctionScheduleId]) -> list[FunctionScheduleResponse]:
        names_by_function: dict[str, set[str]] = defaultdict(set)
        for id_ in ids:
            names_by_function[id_.function_external_id].add(id_.name)
        function_external_ids = [ExternalId(external_id=ext_id) for ext_id in names_by_function]
        functions = FunctionCRUD(self.client, None, None).retrieve(function_external_ids)
        schedules: list[FunctionScheduleResponse] = []
        for func in functions:
            func_external_id = cast(str, func.external_id)
            function_schedules = self.client.tool.functions.schedules.list()
            for schedule in function_schedules:
                if schedule.function_id == func.id:
                    schedule.function_external_id = func_external_id
                    if schedule.name in names_by_function[func_external_id]:
                        schedules.append(schedule)
        return schedules

    def create(self, items: Sequence[FunctionScheduleRequest]) -> list[FunctionScheduleResponse]:
        created_list: list[FunctionScheduleResponse] = []
        function_id_by_external_id = self._get_function_ids_by_external_id(items)

        for item in items:
            id_ = self.get_id(item)
            if id_ not in self.authentication_by_id:
                raise ToolkitRequiredValueError(f"Authentication is missing for schedule {id_!r}")
            client_credentials = self.authentication_by_id[id_]
            try:
                session = self.client.iam.sessions.create(client_credentials, session_type="CLIENT_CREDENTIALS")
            except CogniteAPIError as e:
                if hint := try_find_error(client_credentials):
                    raise ResourceCreationError(f"Failed to create Function Schedule {id_}: {hint}") from e
                raise e

            # Create a new request with the function_id and nonce set
            to_create = FunctionScheduleRequest.model_validate(
                {
                    **item.dump(),
                    "functionId": function_id_by_external_id[id_.function_external_id],
                    "nonce": session.nonce,
                }
            )

            result = self.client.tool.functions.schedules.create([to_create])
            if result:
                created = result[0]
                created.function_external_id = id_.function_external_id
                created_list.append(created)
        return created_list

    def _get_function_ids_by_external_id(self, items: Sequence[FunctionScheduleRequest]) -> dict[str, int]:
        functions_to_lookup = list({item.function_external_id for item in items if item.function_external_id})
        if not functions_to_lookup:
            return {}
        function_ids = self.client.lookup.functions.id(functions_to_lookup)

        if len(function_ids) == len(functions_to_lookup):
            return dict(zip(functions_to_lookup, function_ids))
        # The lookup API is cached, so it is cheap to do individual lookups to find the missing ones.
        lookup_pair = (
            (function_external_id, self.client.lookup.functions.id(function_external_id))
            for function_external_id in functions_to_lookup
        )
        missing_functions = {func for func, func_id in lookup_pair if func_id is None}
        failed_schedules = [self.get_id(item) for item in items if item.function_external_id in missing_functions]
        plural_schedules = "s" if len(failed_schedules) > 1 else ""
        plural_fun = "s" if len(missing_functions) > 1 else ""
        raise ResourceCreationError(
            f"Failed to create function schedule{plural_schedules} {humanize_collection(failed_schedules)}. "
            f"Could not find function{plural_fun} {humanize_collection(missing_functions)!r}"
        )

    def delete(self, ids: Sequence[FunctionScheduleId]) -> int:
        schedules = self.retrieve(ids)
        internal_ids = [InternalId(id=schedule.id) for schedule in schedules if schedule.id]
        self.client.tool.functions.schedules.delete(internal_ids)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[FunctionScheduleResponse]:
        if parent_ids is None:
            for schedules in self.client.tool.functions.schedules.iterate():
                yield from schedules
        else:
            external_ids = [parent_id.external_id for parent_id in parent_ids if isinstance(parent_id, ExternalId)]
            if not external_ids:
                return
            internal_ids = self.client.lookup.functions.id(external_ids)
            for function_id in internal_ids:
                for schedules in self.client.tool.functions.schedules.iterate(function_id=function_id):
                    yield from schedules

    def sensitive_strings(self, item: FunctionScheduleRequest) -> Iterable[str]:
        id_ = self.get_id(item)
        if id_ in self.authentication_by_id:
            yield self.authentication_by_id[id_].client_secret



import sys
from typing import Any

from pydantic import model_validator

from cognite_toolkit._cdf_tk.client._resource_base import RequestItem

from .instance_api import NodeReference

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class PendingInstanceId(RequestItem):
    """A pending instance ID links an asset-centric resource (TimeSeries or FileMetadata)
    to a DM node that will be created later by the syncer service.
    """

    pending_instance_id: NodeReference
    id: int | None = None
    external_id: str | None = None

    @model_validator(mode="after")
    def validate_exactly_one_id(self) -> Self:
        if (self.id is None) == (self.external_id is None):
            raise ValueError("Exactly one of id or external_id must be set.")
        return self

    def __str__(self) -> str:
        id_part = f"id={self.id}" if self.id is not None else f"externalId={self.external_id}"
        return f"PendingInstanceId({id_part})"

    def dump(self, camel_case: bool = True, exclude_extra: bool = False) -> dict[str, Any]:
        return self.model_dump(mode="json", by_alias=camel_case, exclude_unset=True)



from typing import Any

from cognite.client.data_classes.data_modeling import NodeId, ViewId
from pydantic import JsonValue, field_serializer, field_validator

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject


class ChartObject(BaseModelObject, extra="allow"): ...


class UserInfo(ChartObject):
    id: str | None = None
    email: str | None = None
    display_name: str | None = None


class ChartSettings(ChartObject):
    show_y_axis: bool = True
    show_min_max: bool = True
    show_gridlines: bool = True
    merge_units: bool = False


class ThresholdFilter(ChartObject):
    min_unit: str | None = None
    max_unit: str | None = None


class ChartCall(ChartObject):
    id: str | None = None
    hash: int | None = None
    call_id: str | None = None
    call_date: int | None = None
    status: str | None = None


class SubSetting(ChartObject):
    auto_align: bool | None = None


class ChartPosition(ChartObject):
    x: float | None = None
    y: float | None = None


class FlowElement(ChartObject):
    id: str | None = None
    type: str | None = None
    position: ChartPosition | None = None
    data: JsonValue | None = None
    source: str | None = None
    target: str | None = None
    source_handle: str | None = None
    target_handle: str | None = None


class Flow(ChartObject):
    zoom: float | None = None
    elements: list[FlowElement] | None = None
    position: tuple[float | None, float | None] | None = None


class ChartElement(ChartObject):
    id: str | None = None
    type: str | None = None


class ChartSource(ChartElement): ...


class ChartCoreTimeseries(ChartElement):
    node_reference: NodeId | None = None
    view_reference: ViewId | None = None
    display_mode: str | None = None
    color: str | None = None
    created_at: int | None = None
    enabled: bool | None = None
    interpolation: str | None = None
    line_style: str | None = None
    line_weight: float | None = None
    name: str | None = None
    preferred_unit: str | None = None
    range: list[float | None] | None = None

    @field_serializer("node_reference", when_used="always")
    def serialize_node_reference(self, node_reference: NodeId | None) -> dict[str, Any] | None:
        if node_reference:
            return node_reference.dump(include_instance_type=False)
        return None

    @field_serializer("view_reference", when_used="always")
    def serialize_view_reference(self, view_reference: ViewId | None) -> dict[str, Any] | None:
        if view_reference:
            return view_reference.dump(include_type=False)
        return None

    @field_validator("node_reference", mode="before")
    @classmethod
    def validate_node_reference(cls, value: Any) -> NodeId | None:
        if value is None or isinstance(value, NodeId):
            return value
        return NodeId.load(value)

    @field_validator("view_reference", mode="before")
    @classmethod
    def validate_view_reference(cls, value: Any) -> ViewId | None:
        if value is None or isinstance(value, ViewId):
            return value
        return ViewId.load(value)


class ChartTimeseries(ChartElement):
    color: str | None = None
    created_at: int | None = None
    enabled: bool | None = None
    interpolation: str | None = None
    line_style: str | None = None
    line_weight: float | None = None
    name: str | None = None
    preferred_unit: str | None = None
    range: list[float | None] | None = None
    unit: str | None = None
    ts_id: int | None = None
    ts_external_id: str | None = None
    display_mode: str | None = None
    original_unit: str | None = None
    description: str | None = None


class ChartWorkflow(ChartElement):
    version: str | None = None
    name: str | None = None
    color: str | None = None
    enabled: bool | None = None
    line_weight: float | None = None
    line_style: str | None = None
    interpolation: str | None = None
    unit: str | None = None
    preferred_unit: str | None = None
    range: list[float | None] | None = None
    created_at: int | None = None
    settings: SubSetting | None = None
    flow: Flow | None = None
    calls: list[ChartCall] | None = None


class ChartThreshold(ChartElement):
    visible: bool | None = None
    name: str | None = None
    source_id: str | None = None
    upper_limit: float | None = None
    filter: ThresholdFilter | None = None
    calls: list[ChartCall] | None = None


class ChartScheduledCalculation(ChartElement):
    color: str | None = None
    created_at: int | None = None
    description: str | None = None
    enabled: bool | None = None
    interpolation: str | None = None
    line_style: str | None = None
    line_weight: float | None = None
    name: str | None = None
    preferred_unit: str | None = None
    range: list[float | None] | None = None
    unit: str | None = None
    version: str | None = None
    settings: SubSetting | None = None
    flow: Flow | None = None


class ChartData(ChartObject):
    version: int | None = None
    name: str | None = None
    date_from: str | None = None
    date_to: str | None = None
    user_info: UserInfo | None = None
    live_mode: bool | None = None
    time_series_collection: list[ChartTimeseries] | None = None
    core_timeseries_collection: list[ChartCoreTimeseries] | None = None
    workflow_collection: list[ChartWorkflow] | None = None
    source_collection: list[ChartSource] | None = None
    threshold_collection: list[ChartThreshold] | None = None
    scheduled_calculation_collection: list[ChartScheduledCalculation] | None = None
    settings: ChartSettings | None = None



from collections.abc import Iterable, Sequence
from functools import partial
from typing import Literal, TypeVar

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import (
    HTTPClient,
    ItemsSuccessResponse,
    RequestMessage,
    SuccessResponse,
)
from cognite_toolkit._cdf_tk.client.request_classes.filters import ThreeDAssetMappingFilter
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import InternalId, ThreeDModelRevisionId
from cognite_toolkit._cdf_tk.client.resource_classes.three_d import (
    AssetMappingClassicRequest,
    AssetMappingClassicResponse,
    AssetMappingDMRequest,
    AssetMappingDMResponse,
    ThreeDModelClassicRequest,
    ThreeDModelClassicResponse,
    ThreeDRevisionClassicRequest,
    ThreeDRevisionClassicResponse,
)


class ThreeDClassicModelsAPI(CDFResourceAPI[ThreeDModelClassicResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/3d/models", item_limit=1000),
                "delete": Endpoint(method="POST", path="/3d/models/delete", item_limit=1000),
                "update": Endpoint(method="POST", path="/3d/models/update", item_limit=1000),
                "retrieve": Endpoint(method="GET", path="/3d/models/{modelId}", item_limit=1000),
                "list": Endpoint(method="GET", path="/3d/models", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[ThreeDModelClassicResponse]:
        return PagedResponse[ThreeDModelClassicResponse].model_validate_json(response.body)

    def create(self, items: Sequence[ThreeDModelClassicRequest]) -> list[ThreeDModelClassicResponse]:
        """Create 3D models in classic format.

        Args:
            items (Sequence[ThreeDModelClassicRequest]): The 3D model(s) to create.

        Returns:
            list[ThreeDModelClassicResponse]: The created 3D model(s).
        """
        return self._request_item_response(items, "create")

    def retrieve(self, ids: Sequence[InternalId]) -> list[ThreeDModelClassicResponse]:
        """Retrieve 3D models by their IDs.

        Args:
            ids (Sequence[int]): The IDs of the 3D models to retrieve.

        Returns:
            list[ThreeDModelClassicResponse]: The retrieved 3D model(s).
        """
        retrieved: list[ThreeDModelClassicResponse] = []
        endpoint = self._method_endpoint_map["retrieve"]
        for id in ids:
            url = endpoint.path.format(modelId=id.id)
            response = self._http_client.request_single_retries(
                RequestMessage(
                    endpoint_url=self._make_url(url),
                    method=endpoint.method,
                    api_version=self._api_version,
                    disable_gzip=self._disable_gzip,
                )
            )
            result = response.get_success_or_raise()
            retrieved.append(ThreeDModelClassicResponse.model_validate_json(result.body))
        return retrieved

    def update(
        self, items: Sequence[ThreeDModelClassicRequest], mode: Literal["patch", "replace"] = "replace"
    ) -> list[ThreeDModelClassicResponse]:
        """Update 3D models in classic format.

        Args:
            items (Sequence[ThreeDModelClassicRequest]): The 3D model(s) to update.
            mode (Literal["patch", "replace"]): The update mode. "patch" only updates explicitly set fields,
                "replace" replaces all fields.

        Returns:
            list[ThreeDModelClassicResponse]: The updated 3D model(s).
        """
        return self._update(items, mode="replace")

    def delete(self, ids: Sequence[InternalId]) -> None:
        """Delete 3D models by their IDs.

        Args:
            ids (Sequence[int]): The IDs of the 3D models to delete.
        """
        self._request_no_response(ids, "delete")

    @staticmethod
    def _create_list_filter(include_revision_info: bool, published: bool | None) -> dict[str, bool]:
        params = {
            # There is a bug in the API. The parameter includeRevisionInfo is expected to be lower case and not
            # camel case as documented. You get error message: Unrecognized query parameter includeRevisionInfo,
            # did you mean includerevisioninfo?
            "includerevisioninfo": include_revision_info,
        }
        if published is not None:
            params["published"] = published
        return params

    def paginate(
        self,
        published: bool | None = None,
        include_revision_info: bool = False,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[ThreeDModelClassicResponse]:
        params = self._create_list_filter(include_revision_info, published)
        return self._paginate(limit=limit, cursor=cursor, params=params)

    def iterate(
        self,
        published: bool | None = None,
        include_revision_info: bool = False,
        limit: int | None = 100,
        cursor: str | None = None,
    ) -> Iterable[list[ThreeDModelClassicResponse]]:
        params = self._create_list_filter(include_revision_info, published)
        return self._iterate(limit=limit, cursor=cursor, params=params)

    def list(
        self,
        published: bool | None = None,
        include_revision_info: bool = False,
        limit: int | None = 100,
    ) -> list[ThreeDModelClassicResponse]:
        params = self._create_list_filter(include_revision_info, published)
        return self._list(limit=limit, params=params)


class ThreeDClassicRevisionsAPI(CDFResourceAPI[ThreeDRevisionClassicResponse]):
    ENDPOINT = "/3d/models/{modelId}/revisions"

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path=self.ENDPOINT, item_limit=1000),
                "delete": Endpoint(method="POST", path=f"{self.ENDPOINT}/delete", item_limit=1000),
                "update": Endpoint(method="POST", path=f"{self.ENDPOINT}/update", item_limit=1000),
                "list": Endpoint(method="GET", path=self.ENDPOINT, item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[ThreeDRevisionClassicResponse]:
        return PagedResponse[ThreeDRevisionClassicResponse].model_validate_json(response.body)

    def create(self, items: Sequence[ThreeDRevisionClassicRequest]) -> list[ThreeDRevisionClassicResponse]:
        """Create 3D revisions in classic format.

        Items are grouped by model_id and the path is formatted accordingly.

        Args:
            items: The 3D revision(s) to create. Each item must have model_id set.

        Returns:
            The created 3D revision(s).
        """
        results: list[ThreeDRevisionClassicResponse] = []
        for (model_id,), group in self._group_items_by_text_field(items, "model_id").items():
            path = self.ENDPOINT.format(modelId=model_id)
            result = self._request_item_response(group, "create", endpoint=path)
            for item in result:
                item.model_id = int(model_id)
            results.extend(result)
        return results

    def update(
        self,
        items: Sequence[ThreeDRevisionClassicRequest],
        mode: Literal["patch", "replace"] = "replace",
    ) -> list[ThreeDRevisionClassicResponse]:
        """Update 3D revisions in classic format.

        Items are grouped by model_id and the path is formatted accordingly.

        Args:
            items: The 3D revision(s) to update. Each item must have id and model_id set.
            mode: The update mode. "patch" only updates explicitly set fields,
                "replace" replaces all fields.

        Returns:
            The updated 3D revision(s).
        """
        results: list[ThreeDRevisionClassicResponse] = []
        endpoint = self._method_endpoint_map["update"]
        for (model_id,), group in self._group_items_by_text_field(items, "model_id").items():
            path = endpoint.path.format(modelId=model_id)
            for response in self._chunk_requests(
                group, "update", serialization=partial(self._serialize_updates, mode=mode), endpoint_path=path
            ):
                page = self._validate_page_response(response)
                for item in page.items:
                    item.model_id = int(model_id)
                results.extend(page.items)
        return results

    def delete(self, ids: Sequence[ThreeDModelRevisionId]) -> None:
        """Delete 3D revisions by their IDs.

        Args:
            ids: The revisions to delete, identified by a sequence of `ThreeDModelRevisionId` objects.
        """
        endpoint = self._method_endpoint_map["delete"]
        for (model_id,), group in self._group_items_by_text_field(ids, "model_id").items():
            path = endpoint.path.format(modelId=model_id)
            self._request_no_response(group, "delete", endpoint=path)

    @staticmethod
    def _create_list_filter(published: bool | None) -> dict[str, bool]:
        params: dict[str, bool] = {}
        if published is not None:
            params["published"] = published
        return params

    def paginate(
        self,
        model_id: int,
        published: bool | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[ThreeDRevisionClassicResponse]:
        """Fetch a single page of 3D revisions for a model.

        Args:
            model_id: The ID of the model to list revisions for.
            published: Filter based on published status.
            limit: Maximum number of items to return in the page.
            cursor: Cursor for pagination.

        Returns:
            A page of 3D revisions.
        """
        path = self.ENDPOINT.format(modelId=model_id)
        params = self._create_list_filter(published)
        page = self._paginate(limit=limit, cursor=cursor, params=params, endpoint_path=path)
        for item in page.items:
            item.model_id = model_id
        return page

    def iterate(
        self,
        model_id: int,
        published: bool | None = None,
        limit: int | None = 100,
    ) -> Iterable[list[ThreeDRevisionClassicResponse]]:
        """Iterate over all 3D revisions for a model, handling pagination automatically.

        Args:
            model_id: The ID of the model to list revisions for.
            published: Filter based on published status.
            limit: Maximum number of items per page.

        Yields:
            Batches of 3D revisions.
        """
        path = self.ENDPOINT.format(modelId=model_id)
        params = self._create_list_filter(published)
        for items in self._iterate(limit=limit, params=params, endpoint_path=path):
            for item in items:
                item.model_id = model_id
            yield items

    def list(
        self,
        model_id: int,
        published: bool | None = None,
        limit: int | None = 100,
    ) -> list[ThreeDRevisionClassicResponse]:
        """List all 3D revisions for a model.

        Args:
            model_id: The ID of the model to list revisions for.
            published: Filter based on published status.
            limit: Maximum total number of items to return. None means no limit.

        Returns:
            All matching 3D revisions.
        """
        path = self.ENDPOINT.format(modelId=model_id)
        params = self._create_list_filter(published)
        items = self._list(limit=limit, params=params, endpoint_path=path)
        for item in items:
            item.model_id = model_id
        return items


T_RequestMapping = TypeVar("T_RequestMapping", bound=AssetMappingClassicRequest | AssetMappingDMRequest)


class ThreeDClassicAssetMappingAPI(CDFResourceAPI[AssetMappingClassicResponse]):
    ENDPOINT = "/3d/models/{modelId}/revisions/{revisionId}/mappings"

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                # These endpoints are parameterized, so the paths are templates
                "create": Endpoint(method="POST", path=self.ENDPOINT, item_limit=1000),
                "delete": Endpoint(method="POST", path=f"{self.ENDPOINT}/delete", item_limit=1000),
                "list": Endpoint(method="POST", path=f"{self.ENDPOINT}/list", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[AssetMappingClassicResponse]:
        return PagedResponse[AssetMappingClassicResponse].model_validate_json(response.body)

    def create(self, mappings: Sequence[AssetMappingClassicRequest]) -> list[AssetMappingClassicResponse]:
        """Create 3D asset mappings.

        Args:
            mappings (Sequence[AssetMappingClassicRequest]):
                The 3D asset mapping(s) to create.

        Returns:
            list[AssetMappingClassicResponse]: The created 3D asset mapping(s).
        """
        results: list[AssetMappingClassicResponse] = []
        endpoint = self._method_endpoint_map["create"]
        for (model_id, revision_id), group in self._group_items_by_text_field(
            mappings, "model_id", "revision_id"
        ).items():
            path = endpoint.path.format(modelId=model_id, revisionId=revision_id)
            result = self._request_item_response(group, "create", endpoint=path)
            for item in result:
                # We append modelId and revisionId to each item since the API does not return them
                # this is needed to fully populate the AssetMappingResponse data class
                object.__setattr__(item, "model_id", int(model_id))
                object.__setattr__(item, "revision_id", int(revision_id))
            results.extend(result)
        return results

    def delete(self, mappings: Sequence[AssetMappingClassicRequest]) -> None:
        """Delete 3D asset mappings.

        Args:
            mappings (Sequence[AssetMappingClassicRequest]):
                The 3D asset mapping(s) to delete.
        """
        endpoint = self._method_endpoint_map["delete"]
        for (model_id, revision_id), group in self._group_items_by_text_field(
            mappings, "model_id", "revision_id"
        ).items():
            path = endpoint.path.format(modelId=model_id, revisionId=revision_id)
            self._request_no_response(group, "delete", endpoint=path)
        return None

    def paginate(
        self,
        model_id: int,
        revision_id: int,
        filter: ThreeDAssetMappingFilter | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[AssetMappingClassicResponse]:
        endpoint = self._method_endpoint_map["list"]
        path = endpoint.path.format(modelId=model_id, revisionId=revision_id)
        page = self._paginate(
            limit=limit,
            cursor=cursor,
            body={"filter": filter.dump() if filter else None, "getDmsInstances": False},
            endpoint_path=path,
        )
        # Add modelId and revisionId to items since the API does not return them
        for item in page.items:
            object.__setattr__(item, "model_id", model_id)
            object.__setattr__(item, "revision_id", revision_id)
        return page

    def iterate(
        self,
        model_id: int,
        revision_id: int,
        filter: ThreeDAssetMappingFilter | None = None,
        limit: int = 100,
    ) -> Iterable[list[AssetMappingClassicResponse]]:
        endpoint = self._method_endpoint_map["list"]
        path = endpoint.path.format(modelId=model_id, revisionId=revision_id)
        for items in self._iterate(
            body={"filter": filter.dump() if filter else None, "getDmsInstances": False},
            limit=limit,
            endpoint_path=path,
        ):
            # Add modelId and revisionId to items since the API does not return them
            for item in items:
                object.__setattr__(item, "model_id", model_id)
                object.__setattr__(item, "revision_id", revision_id)
            yield items

    def list(
        self,
        model_id: int,
        revision_id: int,
        filter: ThreeDAssetMappingFilter | None = None,
        limit: int | None = 100,
    ) -> list[AssetMappingClassicResponse]:
        endpoint = self._method_endpoint_map["list"]
        path = endpoint.path.format(modelId=model_id, revisionId=revision_id)
        items = self._list(
            body={"filter": filter.dump() if filter else None, "getDmsInstances": False},
            limit=limit,
            endpoint_path=path,
        )
        # Add modelId and revisionId to items since the API does not return them
        for item in items:
            object.__setattr__(item, "model_id", model_id)
            object.__setattr__(item, "revision_id", revision_id)
        return items


class ThreeDDMAssetMappingAPI(CDFResourceAPI[AssetMappingDMResponse]):
    ENDPOINT = "/3d/models/{modelId}/revisions/{revisionId}/mappings"

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                # These endpoints are parameterized, so the paths are templates
                "create": Endpoint(method="POST", path=self.ENDPOINT, item_limit=100),
                "delete": Endpoint(method="POST", path=f"{self.ENDPOINT}/delete", item_limit=100),
                "list": Endpoint(method="POST", path=f"{self.ENDPOINT}/list", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[AssetMappingDMResponse]:
        return PagedResponse[AssetMappingDMResponse].model_validate_json(response.body)

    def create(
        self, mappings: Sequence[AssetMappingDMRequest], object_3d_space: str, cad_node_space: str
    ) -> list[AssetMappingDMResponse]:
        """Create 3D asset mappings in Data Modeling format.

        Args:
            mappings (Sequence[AssetMappingDMRequest]):
                The 3D asset mapping(s) to create
            object_3d_space (str):
                The instance space where the Cognite3DObject are located.
            cad_node_space (str):
                The instance space where the CogniteCADNode are located.
        Returns:
            list[AssetMappingDMResponse]: The created 3D asset mapping(s).
        """
        results: list[AssetMappingDMResponse] = []
        for (model_id, revision_id), group in self._group_items_by_text_field(
            mappings, "model_id", "revision_id"
        ).items():
            path = self.ENDPOINT.format(modelId=model_id, revisionId=revision_id)
            result = self._request_item_response(
                group,
                "create",
                endpoint=path,
                extra_body={
                    "dmsContextualizationConfig": {
                        "object3DSpace": object_3d_space,
                        "cadNodeSpace": cad_node_space,
                    }
                },
            )
            for item in result:
                # We append modelId and revisionId to each item since the API does not return them
                # this is needed to fully populate the AssetMappingDMResponse data class
                object.__setattr__(item, "model_id", int(model_id))
                object.__setattr__(item, "revision_id", int(revision_id))
            results.extend(result)
        return results

    def delete(self, mappings: Sequence[AssetMappingDMRequest], object_3d_space: str, cad_node_space: str) -> None:
        """Delete 3D asset mappings in Data Modeling format.

        Args:
            mappings (Sequence[AssetMappingDMRequest]):
                The 3D asset mapping(s) to delete.
            object_3d_space (str):
                The instance space where the Cognite3DObject are located.
            cad_node_space (str):
                The instance space where the CogniteCADNode are located.
        """
        endpoint = self._method_endpoint_map["delete"]
        for (model_id, revision_id), group in self._group_items_by_text_field(
            mappings, "model_id", "revision_id"
        ).items():
            path = endpoint.path.format(modelId=model_id, revisionId=revision_id)
            self._request_no_response(
                group,
                "delete",
                endpoint=path,
                extra_body={
                    "dmsContextualizationConfig": {
                        "object3DSpace": object_3d_space,
                        "cadNodeSpace": cad_node_space,
                    }
                },
            )
        return None

    def paginate(
        self,
        model_id: int,
        revision_id: int,
        filter: ThreeDAssetMappingFilter | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[AssetMappingDMResponse]:
        endpoint = self._method_endpoint_map["list"]
        path = endpoint.path.format(modelId=model_id, revisionId=revision_id)
        page = self._paginate(
            limit=limit,
            cursor=cursor,
            body={"filter": filter.dump() if filter else None, "getDmsInstances": True},
            endpoint_path=path,
        )
        # Add modelId and revisionId to items since the API does not return them
        for item in page.items:
            object.__setattr__(item, "model_id", model_id)
            object.__setattr__(item, "revision_id", revision_id)
        return page

    def iterate(
        self,
        model_id: int,
        revision_id: int,
        filter: ThreeDAssetMappingFilter | None = None,
        limit: int = 100,
    ) -> Iterable[list[AssetMappingDMResponse]]:
        endpoint = self._method_endpoint_map["list"]
        path = endpoint.path.format(modelId=model_id, revisionId=revision_id)
        for items in self._iterate(
            body={"filter": filter.dump() if filter else None, "getDmsInstances": True}, limit=limit, endpoint_path=path
        ):
            # Add modelId and revisionId to items since the API does not return them
            for item in items:
                object.__setattr__(item, "model_id", model_id)
                object.__setattr__(item, "revision_id", revision_id)
            yield items

    def list(
        self,
        model_id: int,
        revision_id: int,
        filter: ThreeDAssetMappingFilter | None = None,
        limit: int | None = 100,
    ) -> list[AssetMappingDMResponse]:
        endpoint = self._method_endpoint_map["list"]
        path = endpoint.path.format(modelId=model_id, revisionId=revision_id)
        items = self._list(
            body={"filter": filter.dump() if filter else None, "getDmsInstances": True}, limit=limit, endpoint_path=path
        )
        # Add modelId and revisionId to items since the API does not return them
        for item in items:
            object.__setattr__(item, "model_id", model_id)
            object.__setattr__(item, "revision_id", revision_id)
        return items


class ThreeDAPI:
    def __init__(self, http_client: HTTPClient) -> None:
        self.models_classic = ThreeDClassicModelsAPI(http_client)
        self.revisions_classic = ThreeDClassicRevisionsAPI(http_client)
        self.asset_mappings_classic = ThreeDClassicAssetMappingAPI(http_client)
        self.asset_mappings_dm = ThreeDDMAssetMappingAPI(http_client)



from typing import ClassVar

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
    UpdatableRequestResource,
)

from .identifiers import ExternalId


class TransformationSchedule(BaseModelObject):
    external_id: str
    interval: str

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)


class TransformationScheduleRequest(TransformationSchedule, UpdatableRequestResource):
    container_fields: ClassVar[frozenset[str]] = frozenset()
    non_nullable_fields: ClassVar[frozenset[str]] = frozenset({"interval"})
    is_paused: bool | None = None


class TransformationScheduleResponse(TransformationSchedule, ResponseResource[TransformationScheduleRequest]):
    id: int
    created_time: int
    last_updated_time: int
    is_paused: bool

    @classmethod
    def request_cls(cls) -> type[TransformationScheduleRequest]:
        return TransformationScheduleRequest



from collections.abc import Iterable, Sequence
from typing import Any

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.request_classes.filters import ClassicFilter
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId
from cognite_toolkit._cdf_tk.client.resource_classes.label import LabelRequest, LabelResponse


class LabelsAPI(CDFResourceAPI[LabelResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/labels", item_limit=1000),
                "retrieve": Endpoint(method="POST", path="/labels/byids", item_limit=1000),
                "delete": Endpoint(method="POST", path="/labels/delete", item_limit=1000),
                "list": Endpoint(method="POST", path="/labels/list", item_limit=1000),
            },
        )

    def _validate_page_response(self, response: SuccessResponse | ItemsSuccessResponse) -> PagedResponse[LabelResponse]:
        return PagedResponse[LabelResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[ExternalId]:
        return ResponseItems[ExternalId].model_validate_json(response.body)

    def create(self, items: Sequence[LabelRequest]) -> list[LabelResponse]:
        """Create labels in CDF.

        Args:
            items: List of LabelRequest objects to create.
        Returns:
            List of created LabelResponse objects.
        """
        return self._request_item_response(items, "create")

    def retrieve(self, items: Sequence[ExternalId], ignore_unknown_ids: bool = False) -> list[LabelResponse]:
        """Retrieve labels from CDF.

        Args:
            items: List of ExternalId objects to retrieve.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        Returns:
            List of retrieved LabelResponse objects.
        """
        return self._request_item_response(
            items, method="retrieve", extra_body={"ignoreUnknownIds": ignore_unknown_ids}
        )

    def delete(self, items: Sequence[ExternalId]) -> None:
        """Delete labels from CDF.

        Args:
            items: List of ExternalId objects to delete.
        """
        self._request_no_response(items, "delete")

    def paginate(
        self,
        filter: ClassicFilter | None = None,
        name: str | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[LabelResponse]:
        """Iterate over all labels in CDF.

        Args:
            filter: Filter by data set IDs.
            name: Filter by label name (prefix match).
            limit: Maximum number of items to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of LabelResponse objects.
        """
        body: dict[str, Any] = filter.dump() if filter else {}
        if name:
            body["name"] = name

        return self._paginate(
            cursor=cursor,
            limit=limit,
            body=body,
        )

    def iterate(
        self,
        filter: ClassicFilter | None = None,
        limit: int = 100,
    ) -> Iterable[list[LabelResponse]]:
        """Iterate over all labels in CDF.

        Args:
            filter: Filter by data set IDs.
            name: Filter by label name (prefix match).
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of LabelResponse objects.
        """
        body: dict[str, Any] | None = None
        if filter:
            body = {"filter": filter.dump()}
        return self._iterate(
            limit=limit,
            body=body,
        )

    def list(
        self,
        limit: int | None = 100,
    ) -> list[LabelResponse]:
        """List all labels in CDF.

        Returns:
            List of LabelResponse objects.
        """
        return self._list(limit=limit)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from cognite_toolkit._cdf_tk.client._toolkit_client import ToolkitClient
from cognite_toolkit._cdf_tk.commands.build_v2.data_classes._insights import (
    ConsistencyError,
    InsightList,
    ModelSyntaxError,
    Recommendation,
)

if TYPE_CHECKING:
    from cognite.neat._data_model._snapshot import SchemaSnapshot
    from cognite.neat._data_model.models.dms._limits import SchemaLimits
    from cognite.neat._issues import IssueList


class NeatPlugin:
    def __init__(self, client: ToolkitClient) -> None:

        from cognite.neat._client import NeatClient

        self._client = NeatClient(client._config)
        self._cdf_snapshot: SchemaSnapshot | None = None
        self._cdf_limits: SchemaLimits | None = None

    def validate(self, data_model_dir: Path, data_model_file: Path) -> InsightList:
        """Validates a data model using Neat and returns a list of insights.

        Args:
            data_model_dir: The directory containing the data model YAML file.
            data_model_file: The data model YAML file to validate.

        Returns:
            InsightList: A list of insights generated from the validation.
        """

        from cognite.neat._data_model.importers import DMSAPIImporter
        from cognite.neat._data_model.rules.dms import DmsDataModelRulesOrchestrator

        importer = DMSAPIImporter.from_yaml(yaml_file=data_model_dir, data_model_file=data_model_file)
        schema = importer.to_data_model()

        orchestrator = DmsDataModelRulesOrchestrator(
            cdf_snapshot=self.cdf_snapshot,
            limits=self.cdf_limits,
        )
        orchestrator.run(schema)

        return self.issues_to_insights(orchestrator.issues)

    @classmethod
    def issues_to_insights(cls, issues: IssueList) -> InsightList:
        """Converts a list of Neat issues to a Toolkit insight list.

        Args:
            issues: List of Neat issues.

        Returns:
            InsightList: List of Toolkit insights.
        """
        from cognite.neat._issues import ConsistencyError as NeatConsistencyError
        from cognite.neat._issues import ModelSyntaxError as NeatModelSyntaxError
        from cognite.neat._issues import Recommendation as NeatRecommendation

        insights = InsightList()

        for issue in issues:
            if isinstance(issue, NeatModelSyntaxError):
                insights.append(ModelSyntaxError.model_validate(issue.model_dump()))
            elif isinstance(issue, NeatRecommendation):
                insights.append(Recommendation.model_validate(issue.model_dump()))
            elif isinstance(issue, NeatConsistencyError):
                insights.append(ConsistencyError.model_validate(issue.model_dump()))

        return insights

    @property
    def cdf_limits(self) -> SchemaLimits:
        from cognite.neat._data_model.models.dms._limits import SchemaLimits

        if not self._cdf_limits:
            self._cdf_limits = SchemaLimits.from_api_response(self._client.statistics.project())
        return self._cdf_limits

    @property
    def cdf_snapshot(self) -> SchemaSnapshot:
        from cognite.neat._data_model._snapshot import SchemaSnapshot

        if not self._cdf_snapshot:
            self._cdf_snapshot = SchemaSnapshot.fetch_entire_cdf(self._client)
        return self._cdf_snapshot

    @classmethod
    def installed(cls) -> bool:
        """Check if neat is installed"""

        try:
            from cognite import neat

            _ = neat  # silence unused import

            return True
        except ImportError:
            return False



# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import itertools
from collections import defaultdict
from collections.abc import Hashable, Iterable, Sequence
from pathlib import Path
from typing import Any, final

from cognite.client.data_classes.capabilities import (
    Capability,
    RawAcl,
)
from cognite.client.exceptions import CogniteAPIError
from rich import print
from rich.console import Console

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.http_client import ToolkitAPIError
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import NameId, RawDatabaseId, RawTableId
from cognite_toolkit._cdf_tk.client.resource_classes.raw import (
    RAWDatabaseRequest,
    RAWDatabaseResponse,
    RAWTableRequest,
    RAWTableResponse,
)
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceContainerCRUD, ResourceCRUD
from cognite_toolkit._cdf_tk.resource_classes import DatabaseYAML, TableYAML

from .auth import GroupAllScopedCRUD


@final
class RawDatabaseCRUD(ResourceContainerCRUD[RawDatabaseId, RAWDatabaseRequest, RAWDatabaseResponse]):
    item_name = "raw tables"
    folder_name = "raw"
    resource_cls = RAWDatabaseResponse
    resource_write_cls = RAWDatabaseRequest
    kind = "Database"
    yaml_cls = DatabaseYAML
    dependencies = frozenset({GroupAllScopedCRUD})
    support_update = False
    _doc_url = "Raw/operation/createDBs"

    def __init__(self, client: ToolkitClient, build_dir: Path, console: Console | None):
        super().__init__(client, build_dir, console)
        self._loaded_db_names: set[str] = set()

    @property
    def display_name(self) -> str:
        return "raw databases"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[RAWDatabaseRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [RawAcl.Action.Read, RawAcl.Action.List]
            if read_only
            else [RawAcl.Action.Read, RawAcl.Action.Write, RawAcl.Action.List]
        )

        scope: RawAcl.Scope.All | RawAcl.Scope.Table = RawAcl.Scope.All()  # type: ignore[valid-type]
        if items:
            tables_by_database: dict[str, list[str]] = {}
            for item in items:
                tables_by_database[item.name] = []

            scope = RawAcl.Scope.Table(dict(tables_by_database)) if tables_by_database else RawAcl.Scope.All()

        return RawAcl(actions, scope)

    @classmethod
    def get_id(cls, item: RAWDatabaseResponse | RAWDatabaseRequest | dict) -> RawDatabaseId:
        if isinstance(item, dict):
            return RawDatabaseId.model_validate(item)
        return RawDatabaseId(name=item.name)

    @classmethod
    def dump_id(cls, id: RawDatabaseId) -> dict[str, Any]:
        return {"dbName": id.name}

    def dump_resource(self, resource: RAWDatabaseResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        return {"dbName": resource.name}

    def create(self, items: Sequence[RAWDatabaseRequest]) -> list[RAWDatabaseResponse]:
        return self.client.tool.raw.databases.create(items)

    def retrieve(self, ids: Sequence[RawDatabaseId]) -> list[RAWDatabaseResponse]:
        database_list = self.client.tool.raw.databases.list(limit=None)
        target_dbs = {db.name for db in ids}
        return [db for db in database_list if db.name in target_dbs]

    def delete(self, ids: Sequence[RawDatabaseId]) -> int:
        ids_list = list(ids)
        try:
            self.client.tool.raw.databases.delete(ids_list)
        except ToolkitAPIError as e:
            # Bug in API, missing is returned as failed
            if e.missing and (remaining := [db for db in ids_list if db.name not in e.missing]):
                self.client.tool.raw.databases.delete(remaining)
            elif e.code == 404 and "not found" in e.message and "database" in e.message:
                return 0
            else:
                raise e
        return len(ids_list)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[RAWDatabaseResponse]:
        for databases in self.client.tool.raw.databases.iterate():
            yield from databases

    def count(self, ids: Sequence[RawDatabaseId]) -> int:
        nr_of_tables = 0
        for db_name, raw_tables in itertools.groupby(sorted(ids, key=lambda x: x.name), key=lambda x: x.name):
            try:
                tables = self.client.tool.raw.tables.list(db_name=db_name, limit=None)
            except CogniteAPIError as e:
                if db_name in {item.get("name") for item in e.missing or []}:
                    continue
                raise e
            nr_of_tables += len(tables)
        return nr_of_tables

    def drop_data(self, ids: Sequence[RawDatabaseId]) -> int:
        nr_of_tables = 0
        for db_name, raw_tables in itertools.groupby(sorted(ids, key=lambda x: x.name), key=lambda x: x.name):
            try:
                existing = self.client.tool.raw.tables.list(db_name=db_name, limit=None)
            except ToolkitAPIError as e:
                if db_name in {item.get("name") for item in e.missing or []}:
                    continue
                raise e
            if existing:
                self.client.tool.raw.tables.delete([table.as_id() for table in existing])
                nr_of_tables += len(existing)
        return nr_of_tables


@final
class RawTableCRUD(ResourceContainerCRUD[RawTableId, RAWTableRequest, RAWTableResponse]):
    item_name = "raw rows"
    folder_name = "raw"
    resource_cls = RAWTableResponse
    resource_write_cls = RAWTableRequest
    kind = "Table"
    yaml_cls = TableYAML
    support_update = False
    dependencies = frozenset({RawDatabaseCRUD, GroupAllScopedCRUD})
    _doc_url = "Raw/operation/createTables"
    parent_resource = frozenset({RawDatabaseCRUD})

    def __init__(self, client: ToolkitClient, build_dir: Path, console: Console | None):
        super().__init__(client, build_dir, console)
        self._printed_warning = False

    @property
    def display_name(self) -> str:
        return "raw tables"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[RAWTableRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [RawAcl.Action.Read, RawAcl.Action.List]
            if read_only
            else [RawAcl.Action.Read, RawAcl.Action.Write, RawAcl.Action.List]
        )

        scope: RawAcl.Scope.All | RawAcl.Scope.Table = RawAcl.Scope.All()  # type: ignore[valid-type]
        if items:
            tables_by_database = defaultdict(list)
            for item in items:
                tables_by_database[item.db_name].append(item.name)

            scope = RawAcl.Scope.Table(dict(tables_by_database)) if tables_by_database else RawAcl.Scope.All()

        return RawAcl(actions, scope)

    @classmethod
    def get_id(cls, item: RAWTableResponse | RAWTableRequest | dict) -> RawTableId:
        if isinstance(item, dict):
            if missing := tuple(k for k in {"dbName", "tableName"} if k not in item):
                # We need to raise a KeyError with all missing keys to get the correct error message.
                raise KeyError(*missing)
            return RawTableId(db_name=item["dbName"], name=item["tableName"])
        return RawTableId(db_name=item.db_name, name=item.name)

    @classmethod
    def dump_id(cls, id: RawTableId) -> dict[str, Any]:
        return {"dbName": id.db_name, "tableName": id.name}

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "dbName" in item:
            yield RawDatabaseCRUD, RawDatabaseId(name=item["dbName"])

    def dump_resource(self, resource: RAWTableResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        return {"dbName": resource.db_name, "tableName": resource.name}

    def create(self, items: Sequence[RAWTableRequest]) -> list[RAWTableResponse]:
        return self.client.tool.raw.tables.create(items)

    def retrieve(self, ids: Sequence[RawTableId]) -> list[RAWTableResponse]:
        retrieved: list[RAWTableResponse] = []
        for db_name, raw_tables in itertools.groupby(sorted(ids, key=lambda x: x.db_name), key=lambda x: x.db_name):
            expected_tables = {table.name for table in raw_tables}
            try:
                tables = self.client.tool.raw.tables.list(db_name=db_name, limit=None)
            except ToolkitAPIError as e:
                if db_name in {item.get("name") for item in e.missing or []}:
                    continue
                raise e
            retrieved.extend(table for table in tables if table.name in expected_tables)
        return retrieved

    def delete(self, ids: Sequence[RawTableId]) -> int:
        count = 0
        for db_name, raw_tables in itertools.groupby(sorted(ids, key=lambda x: x.db_name), key=lambda x: x.db_name):
            tables_to_delete = [table for table in raw_tables if table.name]
            if tables_to_delete:
                try:
                    self.client.tool.raw.tables.delete(tables_to_delete)
                except ToolkitAPIError as e:
                    if e.code != 404:
                        raise e
                    missing = {item.get("name") for item in (e.missing or [])}
                    if "not found" in e.message and "database" in e.message:
                        continue
                    elif remaining := [t for t in tables_to_delete if t.name not in missing]:
                        self.client.tool.raw.tables.delete(remaining)
                    elif not remaining:
                        # Table does not exist.
                        continue
                    else:
                        raise e
                count += len(tables_to_delete)
        return count

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[RAWTableResponse]:
        if parent_ids is None:
            dbs = self.client.tool.raw.databases.list(limit=None)
            parent_ids = [RawDatabaseId(name=db.name) for db in dbs]
        for parent_id in parent_ids:
            if not isinstance(parent_id, NameId):
                continue
            for tables in self.client.tool.raw.tables.iterate(db_name=parent_id.name):
                yield from tables

    def count(self, ids: Sequence[RawTableId]) -> int:
        if not self._printed_warning:
            print("  [bold green]INFO:[/] Raw rows do not support count (there is no aggregation method).")
            self._printed_warning = True
        return -1

    def drop_data(self, ids: Sequence[RawTableId]) -> int:
        for db_name, raw_tables in itertools.groupby(sorted(ids, key=lambda x: x.db_name), key=lambda x: x.db_name):
            try:
                existing_tables = self.client.tool.raw.tables.list(db_name=db_name, limit=None)
                existing_names = {table.name for table in existing_tables}
            except CogniteAPIError as e:
                if db_name in {item.get("name") for item in e.missing or []}:
                    continue
                raise e
            tables_to_delete = [
                RAWTableResponse(db_name=db_name, name=table.name)
                for table in raw_tables
                if table.name in existing_names
            ]
            if tables_to_delete:
                self.client.tool.raw.tables.delete([table.as_id() for table in tables_to_delete])
        return -1



from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import (
    BtreeIndex,
    ContainerPropertyDefinition,
    ContainerRequest,
    DataModelRequest,
    DirectNodeRelation,
    EnumProperty,
    EnumValue,
    Int64Property,
    JSONProperty,
    SpaceRequest,
    TextProperty,
    UniquenessConstraintDefinition,
    ViewCorePropertyRequest,
    ViewRequest,
)

SPACE = SpaceRequest(
    space="cognite_migration",
    description="Space for the asset-centric to data modeling migration",
    name="cdf_migration",
)
COGNITE_MIGRATION_SPACE_ID = SPACE.space

RESOURCE_VIEW_MAPPING = ContainerRequest(
    space=SPACE.space,
    external_id="ResourceViewMapping",
    used_for="node",
    properties={
        "resourceType": ContainerPropertyDefinition(
            type=TextProperty(max_text_size=255),
            nullable=False,
        ),
        "viewId": ContainerPropertyDefinition(
            type=JSONProperty(),
            nullable=False,
        ),
        "propertyMapping": ContainerPropertyDefinition(
            type=JSONProperty(),
            nullable=False,
        ),
    },
)

INSTANCE_SOURCE_CONTAINER = ContainerRequest(
    space=SPACE.space,
    external_id="InstanceSource",
    used_for="node",
    properties={
        "resourceType": ContainerPropertyDefinition(
            type=EnumProperty(
                values={
                    "timeseries": EnumValue(),
                    "asset": EnumValue(),
                    "file": EnumValue(),
                    "event": EnumValue(),
                    "sequence": EnumValue(),
                }
            ),
            nullable=False,
        ),
        "id": ContainerPropertyDefinition(
            type=Int64Property(),
            nullable=False,
        ),
        "dataSetId": ContainerPropertyDefinition(
            type=Int64Property(),
            nullable=True,
        ),
        "classicExternalId": ContainerPropertyDefinition(
            type=TextProperty(),
            nullable=True,
        ),
        "preferredConsumerViewId": ContainerPropertyDefinition(
            type=JSONProperty(),
            nullable=True,
        ),
        "resourceViewMapping": ContainerPropertyDefinition(
            type=DirectNodeRelation(container=RESOURCE_VIEW_MAPPING.as_id()), nullable=True
        ),
    },
    indexes={
        "id": BtreeIndex(properties=["id"], cursorable=True),
        "resourceType": BtreeIndex(properties=["resourceType", "id"], cursorable=False),
    },
)

CREATED_SOURCE_SYSTEM = ContainerRequest(
    space=SPACE.space,
    external_id="CreatedSourceSystem",
    used_for="node",
    properties={
        "source": ContainerPropertyDefinition(
            type=TextProperty(max_text_size=128),
            nullable=False,
        )
    },
    constraints={"sourceUnique": UniquenessConstraintDefinition(properties=["source"])},
    indexes={
        "source": BtreeIndex(properties=["source"], cursorable=True),
    },
)

SPACE_SOURCE = ContainerRequest(
    space=SPACE.space,
    external_id="SpaceSource",
    used_for="node",
    properties={
        "instanceSpace": ContainerPropertyDefinition(
            type=TextProperty(max_text_size=64),
            nullable=False,
        ),
        "dataSetId": ContainerPropertyDefinition(
            type=Int64Property(),
            nullable=False,
        ),
        "dataSetExternalId": ContainerPropertyDefinition(
            type=TextProperty(max_text_size=256),
            nullable=True,
        ),
    },
    indexes={
        "space": BtreeIndex(properties=["instanceSpace"], cursorable=True),
        "dataSetId": BtreeIndex(properties=["dataSetId"], cursorable=True),
        "dataSetExternalId": BtreeIndex(properties=["dataSetExternalId"], cursorable=True),
    },
    constraints={
        "dataSetIdUnique": UniquenessConstraintDefinition(properties=["dataSetId"]),
    },
)

CONTAINERS = [RESOURCE_VIEW_MAPPING, INSTANCE_SOURCE_CONTAINER, CREATED_SOURCE_SYSTEM, SPACE_SOURCE]

RESOURCE_VIEW_MAPPING_VIEW = ViewRequest(
    space=SPACE.space,
    external_id="ResourceViewMapping",
    version="v1",
    name="ResourceViewMapping",
    description="The mapping from asset-centric resources to data modeling view.",
    properties={
        "resourceType": ViewCorePropertyRequest(
            container=RESOURCE_VIEW_MAPPING.as_id(),
            container_property_identifier="resourceType",
        ),
        "viewId": ViewCorePropertyRequest(
            container=RESOURCE_VIEW_MAPPING.as_id(),
            container_property_identifier="viewId",
        ),
        "propertyMapping": ViewCorePropertyRequest(
            container=RESOURCE_VIEW_MAPPING.as_id(),
            container_property_identifier="propertyMapping",
        ),
    },
)

INSTANCE_SOURCE_VIEW = ViewRequest(
    space=SPACE.space,
    external_id="InstanceSource",
    version="v1",
    name="InstanceSource",
    description="The source of the instance in asset-centric resources.",
    properties={
        "resourceType": ViewCorePropertyRequest(
            container=INSTANCE_SOURCE_CONTAINER.as_id(),
            container_property_identifier="resourceType",
        ),
        "id": ViewCorePropertyRequest(
            container=INSTANCE_SOURCE_CONTAINER.as_id(),
            container_property_identifier="id",
        ),
        "dataSetId": ViewCorePropertyRequest(
            container=INSTANCE_SOURCE_CONTAINER.as_id(),
            container_property_identifier="dataSetId",
        ),
        "classicExternalId": ViewCorePropertyRequest(
            container=INSTANCE_SOURCE_CONTAINER.as_id(),
            container_property_identifier="classicExternalId",
        ),
        "preferredConsumerViewId": ViewCorePropertyRequest(
            container=INSTANCE_SOURCE_CONTAINER.as_id(),
            container_property_identifier="preferredConsumerViewId",
        ),
        "resourceViewMapping": ViewCorePropertyRequest(
            container=INSTANCE_SOURCE_CONTAINER.as_id(),
            container_property_identifier="resourceViewMapping",
            source=RESOURCE_VIEW_MAPPING_VIEW.as_id(),
        ),
    },
)

CREATED_SOURCE_SYSTEM_VIEW = ViewRequest(
    space=SPACE.space,
    external_id="CreatedSourceSystem",
    version="v1",
    name="CreatedSourceSystem",
    description="The source string the SourceSystem was created from.",
    properties={
        "source": ViewCorePropertyRequest(
            container=CREATED_SOURCE_SYSTEM.as_id(),
            container_property_identifier="source",
        ),
    },
)

SPACE_SOURCE_VIEW = ViewRequest(
    space=SPACE.space,
    external_id="SpaceSource",
    version="v1",
    name="SpaceSource",
    description="The mapping from CDF spaces to data sets.",
    properties={
        "instanceSpace": ViewCorePropertyRequest(
            container=SPACE_SOURCE.as_id(),
            container_property_identifier="instanceSpace",
            description="The identifier of the created instances space.",
        ),
        "dataSetId": ViewCorePropertyRequest(
            container=SPACE_SOURCE.as_id(),
            container_property_identifier="dataSetId",
            description="The dataSetId the space was created from.",
        ),
        "dataSetExternalId": ViewCorePropertyRequest(
            container=SPACE_SOURCE.as_id(),
            container_property_identifier="dataSetExternalId",
            description="The externalId of the dataSet (if present) the space was created from.",
        ),
    },
)

INSTANCE_SOURCE_VIEW_ID = INSTANCE_SOURCE_VIEW.as_id()
RESOURCE_VIEW_MAPPING_VIEW_ID = RESOURCE_VIEW_MAPPING_VIEW.as_id()
CREATED_SOURCE_SYSTEM_VIEW_ID = CREATED_SOURCE_SYSTEM_VIEW.as_id()
SPACE_SOURCE_VIEW_ID = SPACE_SOURCE_VIEW.as_id()

VIEWS = [RESOURCE_VIEW_MAPPING_VIEW, INSTANCE_SOURCE_VIEW, CREATED_SOURCE_SYSTEM_VIEW, SPACE_SOURCE_VIEW]

COGNITE_MIGRATION_MODEL = DataModelRequest(
    space=SPACE.space,
    external_id="CogniteMigration",
    version="v1",
    name="CDF Migration Model",
    description="Data model for migrating asset-centric resources to data modeling resources in CDF.",
    views=[INSTANCE_SOURCE_VIEW_ID, RESOURCE_VIEW_MAPPING_VIEW_ID, CREATED_SOURCE_SYSTEM_VIEW_ID, SPACE_SOURCE_VIEW_ID],
)

MODEL_ID = COGNITE_MIGRATION_MODEL.as_id()



from ._capability import RobotCapability, RobotCapabilityRequest, RobotCapabilityResponse
from ._common import MapType, Point3D, Quaternion, RobotType, Transform
from ._data_post_processing import (
    RobotDataPostProcessing,
    RobotDataPostProcessingRequest,
    RobotDataPostProcessingResponse,
)
from ._frame import RobotFrame, RobotFrameRequest, RobotFrameResponse
from ._location import RobotLocation, RobotLocationRequest, RobotLocationResponse
from ._map import RobotMap, RobotMapRequest, RobotMapResponse
from ._robot import Robot, RobotRequest, RobotResponse

__all__ = [
    "MapType",
    "Point3D",
    "Quaternion",
    "Robot",
    "RobotCapability",
    "RobotCapabilityRequest",
    "RobotCapabilityResponse",
    "RobotDataPostProcessing",
    "RobotDataPostProcessingRequest",
    "RobotDataPostProcessingResponse",
    "RobotFrame",
    "RobotFrameRequest",
    "RobotFrameResponse",
    "RobotLocation",
    "RobotLocationRequest",
    "RobotLocationResponse",
    "RobotMap",
    "RobotMapRequest",
    "RobotMapResponse",
    "RobotRequest",
    "RobotResponse",
    "RobotType",
    "Transform",
]



from cognite_toolkit._cdf_tk.client.http_client import HTTPClient

from .hosted_extractor_destinations import HostedExtractorDestinationsAPI
from .hosted_extractor_jobs import HostedExtractorJobsAPI
from .hosted_extractor_mappings import HostedExtractorMappingsAPI
from .hosted_extractor_sources import HostedExtractorSourcesAPI


class HostedExtractorsAPI:
    """API for Hosted Extractors resources.

    This class groups all hosted extractor related APIs:
    - sources: Manage hosted extractor sources (MQTT, Kafka, EventHub, REST)
    - jobs: Manage hosted extractor jobs
    - destinations: Manage hosted extractor destinations
    - mappings: Manage hosted extractor mappings
    """

    def __init__(self, http_client: HTTPClient) -> None:
        self.sources = HostedExtractorSourcesAPI(http_client)
        self.jobs = HostedExtractorJobsAPI(http_client)
        self.destinations = HostedExtractorDestinationsAPI(http_client)
        self.mappings = HostedExtractorMappingsAPI(http_client)



from pydantic import Field

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    RequestResource,
    ResponseResource,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExtractionPipelineConfigId


class ExtractionPipelineConfig(BaseModelObject):
    """Base class for extraction pipeline configuration revisions."""

    external_id: str
    config: str | None = None
    description: str | None = None


class ExtractionPipelineConfigRequest(ExtractionPipelineConfig, RequestResource):
    """Request body for creating a new extraction pipeline configuration revision.

    Attributes:
        external_id: External ID of the extraction pipeline this configuration revision belongs to.
        config: Configuration content.
        description: A description of this configuration revision.
    """

    revision: int | None = Field(None, exclude=True)

    def as_id(self) -> ExtractionPipelineConfigId:
        return ExtractionPipelineConfigId(external_id=self.external_id, revision=self.revision)


class ExtractionPipelineConfigResponse(ExtractionPipelineConfig, ResponseResource[ExtractionPipelineConfigRequest]):
    """Response for an extraction pipeline configuration revision.

    Attributes:
        external_id: External ID of the extraction pipeline.
        config: Configuration content.
        revision: The revision number.
        created_time: Timestamp when this revision was created, in milliseconds since epoch.
        description: A description of this configuration revision.
    """

    revision: int
    created_time: int

    @classmethod
    def request_cls(cls) -> type[ExtractionPipelineConfigRequest]:
        return ExtractionPipelineConfigRequest



from collections.abc import Iterable, Sequence
from typing import ClassVar

from cognite_toolkit._cdf_tk.client.http_client import HTTPClient
from cognite_toolkit._cdf_tk.client.http_client._item_classes import ItemsRequest, ItemsResultList
from cognite_toolkit._cdf_tk.client.resource_classes.records import RecordRequest, RecordResponse
from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal

from ._base import Page, UploadableStorageIO, UploadItem
from .selectors import RecordContainerSelector


class RecordIO(UploadableStorageIO[RecordContainerSelector, RecordResponse, RecordRequest]):  # pyright: ignore[reportInvalidTypeArguments]
    KIND = "Records"
    SUPPORTED_READ_FORMATS: ClassVar[frozenset[str]] = frozenset({".ndjson"})
    UPLOAD_ENDPOINT = "/streams/{streamId}/records"
    CHUNK_SIZE = 1000
    BASE_SELECTOR = RecordContainerSelector

    def as_id(self, item: RecordResponse) -> str:
        return f"{item.space}:{item.external_id}"

    def count(self, selector: RecordContainerSelector) -> int | None:
        return None

    def stream_data(self, selector: RecordContainerSelector, limit: int | None = None) -> Iterable[Page]:
        raise NotImplementedError()

    def data_to_json_chunk(
        self, data_chunk: Sequence[RecordResponse], selector: RecordContainerSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        raise NotImplementedError()

    def json_to_resource(self, item_json: dict[str, JsonVal]) -> RecordRequest:
        return RecordRequest.model_validate(item_json)

    def upload_items(
        self,
        data_chunk: Sequence[UploadItem[RecordRequest]],  # pyright: ignore[reportInvalidTypeArguments]
        http_client: HTTPClient,
        selector: RecordContainerSelector | None = None,
    ) -> ItemsResultList:
        if selector is None:
            raise ToolkitValueError("Selector must be provided for RecordIO upload_items")
        url = self.UPLOAD_ENDPOINT.format(streamId=selector.stream.external_id)
        config = http_client.config
        return http_client.request_items_retries(
            message=ItemsRequest(
                endpoint_url=config.create_api_url(url),
                method="POST",
                items=data_chunk,
            )
        )



from typing import ClassVar

from pydantic import JsonValue

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
    UpdatableRequestResource,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId


class RobotCapability(BaseModelObject):
    """Robot capabilities define what actions that robots can execute, including data capture (PTZ, PTZ-IR, 360)
    and behaviors (e.g., docking).

    Args:
        name: RobotCapability name.
        external_id: RobotCapability external id. Must be unique for the resource type.
        method: RobotCapability method. The method is used to call the right functionality on the robot.
        input_schema: Schema that defines what inputs are needed for the action.
            The input are values that configure the action, e.g pan, tilt and zoom values.
        data_handling_schema: Schema that defines how the data from a RobotCapability should be handled,
            including upload instructions.
        description: Description of RobotCapability. Textual description of the RobotCapability.
    """

    external_id: str
    name: str
    method: str
    description: str | None = None

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)


class RobotCapabilityRequest(RobotCapability, UpdatableRequestResource):
    """Request resource for creating or updating a RobotCapability."""

    non_nullable_fields: ClassVar[frozenset[str]] = frozenset({"input_schema", "data_handling_schema"})
    input_schema: dict[str, JsonValue] | None = None
    data_handling_schema: dict[str, JsonValue] | None = None


class RobotCapabilityResponse(RobotCapability, ResponseResource[RobotCapabilityRequest]):
    """Response resource for a RobotCapability."""

    # The response always has input_schema and data_handling_schema
    input_schema: dict[str, JsonValue]
    data_handling_schema: dict[str, JsonValue]

    @classmethod
    def request_cls(cls) -> type[RobotCapabilityRequest]:
        return RobotCapabilityRequest



"""Spaces API for managing CDF data modeling spaces.

Based on the API specification at:
https://api-docs.cognite.com/20230101/tag/Spaces/operation/ApplySpaces
"""

from collections.abc import Iterable, Sequence

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, Endpoint, PagedResponse
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, SuccessResponse
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import SpaceReference, SpaceRequest, SpaceResponse


class SpacesAPI(CDFResourceAPI[SpaceResponse]):
    """API for managing CDF data modeling spaces.

    Spaces use an apply/upsert pattern for create and update operations.
    """

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "upsert": Endpoint(method="POST", path="/models/spaces", item_limit=100),
                "retrieve": Endpoint(method="POST", path="/models/spaces/byids", item_limit=100),
                "delete": Endpoint(method="POST", path="/models/spaces/delete", item_limit=100),
                "list": Endpoint(method="GET", path="/models/spaces", item_limit=1000),
            },
        )

    def _validate_page_response(self, response: SuccessResponse | ItemsSuccessResponse) -> PagedResponse[SpaceResponse]:
        return PagedResponse[SpaceResponse].model_validate_json(response.body)

    def create(self, items: Sequence[SpaceRequest]) -> list[SpaceResponse]:
        """Apply (create or update) spaces in CDF.

        Args:
            items: List of SpaceRequest objects to apply.

        Returns:
            List of applied SpaceResponse objects.
        """
        return self._request_item_response(items, "upsert")

    def retrieve(self, items: Sequence[SpaceReference]) -> list[SpaceResponse]:
        """Retrieve spaces from CDF.

        Args:
            items: List of SpaceReference objects to retrieve.

        Returns:
            List of retrieved SpaceResponse objects.
        """
        return self._request_item_response(items, method="retrieve")

    def delete(self, items: Sequence[SpaceReference]) -> None:
        """Delete spaces from CDF.

        Args:
            items: List of SpaceReference objects to delete.
        """
        self._request_no_response(items, "delete")

    def paginate(
        self,
        include_global: bool = False,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[SpaceResponse]:
        """Get a page of spaces from CDF.

        Args:
            include_global: Whether to include global spaces.
            limit: Maximum number of spaces to return.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of SpaceResponse objects.
        """
        return self._paginate(
            cursor=cursor,
            limit=limit,
            params={"includeGlobal": include_global},
        )

    def iterate(
        self,
        include_global: bool = False,
        limit: int | None = None,
    ) -> Iterable[list[SpaceResponse]]:
        """Iterate over all spaces in CDF.

        Args:
            include_global: Whether to include global spaces.
            limit: Maximum total number of spaces to return.

        Returns:
            Iterable of lists of SpaceResponse objects.
        """
        return self._iterate(
            limit=limit,
            params={"includeGlobal": include_global},
        )

    def list(self, include_global: bool = False, limit: int | None = None) -> list[SpaceResponse]:
        """List all spaces in CDF.

        Args:
            include_global: Whether to include global spaces.
            limit: Maximum total number of spaces to return.

        Returns:
            List of SpaceResponse objects.
        """
        return self._list(limit=limit, params={"includeGlobal": include_global})



"""Data classes for handling item-based requests and responses in the Cognite Toolkit HTTP client."""

import gzip
from collections import UserList
from collections.abc import Sequence
from typing import Any

from cognite.client import global_config
from pydantic import BaseModel, ConfigDict, Field, JsonValue

from cognite_toolkit._cdf_tk.client._resource_base import RequestItem
from cognite_toolkit._cdf_tk.client.http_client._data_classes import (
    _BODY_SERIALIZER,
    BaseRequestMessage,
    ErrorDetails,
)
from cognite_toolkit._cdf_tk.client.http_client._exception import ToolkitAPIError
from cognite_toolkit._cdf_tk.client.http_client._tracker import ItemsRequestTracker


class ItemsResultMessage(BaseModel):
    ids: list[str]


class ItemsFailedRequest(ItemsResultMessage):
    error_message: str


class ItemsSuccessResponse(ItemsResultMessage):
    status_code: int
    body: str
    content: bytes


class ItemsFailedResponse(ItemsResultMessage):
    status_code: int
    error: ErrorDetails
    body: str


def _set_default_tracker(data: dict[str, Any]) -> ItemsRequestTracker:
    if "tracker" not in data or data["tracker"] is None:
        return ItemsRequestTracker(data.get("max_failures_before_abort", 50))
    return data["tracker"]


class ItemsRequest(BaseRequestMessage):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    items: Sequence[RequestItem]
    extra_body_fields: dict[str, JsonValue] | None = None
    max_failures_before_abort: int = 50
    tracker: ItemsRequestTracker = Field(init=False, default_factory=_set_default_tracker, exclude=True)

    @property
    def content(self) -> str | bytes | None:
        body: dict[str, JsonValue] = {"items": [item.dump(camel_case=True) for item in self.items]}
        if self.extra_body_fields:
            body.update(self.extra_body_fields)
        res = _BODY_SERIALIZER.dump_json(body)
        if not global_config.disable_gzip and not self.disable_gzip and isinstance(res, bytes):
            return gzip.compress(res)
        return res

    def split(self, status_attempts: int) -> list["ItemsRequest"]:
        """Split the request into multiple requests with a single item each."""
        mid = len(self.items) // 2
        if mid == 0:
            return [self]
        self.tracker.register_failure()
        messages: list[ItemsRequest] = []
        for part in (self.items[:mid], self.items[mid:]):
            new_request = self.model_copy(update={"items": part, "status_attempt": status_attempts})
            new_request.tracker = self.tracker
            messages.append(new_request)
        return messages


class ItemResponse(BaseModel):
    items: list[dict[str, JsonValue]]


class ItemsResultList(UserList[ItemsResultMessage]):
    def __init__(self, collection: Sequence[ItemsResultMessage] | None = None) -> None:
        super().__init__(collection or [])

    def raise_for_status(self) -> None:
        """Raises an exception if any response in the list indicates a failure."""
        failed_responses = [resp for resp in self.data if isinstance(resp, ItemsFailedResponse)]
        failed_requests = [resp for resp in self.data if isinstance(resp, ItemsFailedRequest)]
        if not failed_responses and not failed_requests:
            return
        error_messages = "; ".join(f"Status {err.status_code}: {err.error.message}" for err in failed_responses)
        if failed_requests:
            if error_messages:
                error_messages += "; "
            error_messages += "; ".join(f"Request error: {err.error_message}" for err in failed_requests)
        raise ToolkitAPIError(f"One or more requests failed: {error_messages}")

    @property
    def has_failed(self) -> bool:
        """Indicates whether any response in the list indicates a failure.

        Returns:
            bool: True if there are any failed responses or requests, False otherwise.
        """
        for resp in self.data:
            if isinstance(resp, ItemsFailedResponse | ItemsFailedRequest):
                return True
        return False

    def get_items(self) -> list[dict[str, JsonValue]]:
        """Get the items from all successful responses."""
        items: list[dict[str, JsonValue]] = []
        for resp in self.data:
            if isinstance(resp, ItemsSuccessResponse):
                body_json = ItemResponse.model_validate_json(resp.body)
                items.extend(body_json.items)
        return items



from collections.abc import Hashable, Iterable, Sequence
from typing import Any, final

from cognite.client.data_classes import capabilities
from cognite.client.data_classes.capabilities import Capability

from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import (
    ExternalId,
    InternalId,
    NameId,
    ThreeDModelRevisionId,
)
from cognite_toolkit._cdf_tk.client.resource_classes.three_d import (
    ThreeDModelClassicRequest,
    ThreeDModelClassicResponse,
)
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceContainerCRUD, ResourceCRUD
from cognite_toolkit._cdf_tk.resource_classes import ThreeDModelYAML
from cognite_toolkit._cdf_tk.utils import sanitize_filename

from .data_organization import DataSetsCRUD


@final
class ThreeDModelCRUD(ResourceContainerCRUD[NameId, ThreeDModelClassicRequest, ThreeDModelClassicResponse]):
    folder_name = "3dmodels"
    resource_cls = ThreeDModelClassicResponse
    resource_write_cls = ThreeDModelClassicRequest
    kind = "3DModel"
    yaml_cls = ThreeDModelYAML
    dependencies = frozenset({DataSetsCRUD})
    _doc_url = "3D-Models/operation/create3DModels"
    item_name = "revisions"

    @property
    def display_name(self) -> str:
        return "3D models"

    @classmethod
    def get_id(cls, item: ThreeDModelClassicRequest | ThreeDModelClassicResponse | dict) -> NameId:
        if isinstance(item, dict):
            return NameId(name=item["name"])
        if not item.name:
            raise KeyError("3DModel must have name")
        return NameId(name=item.name)

    @classmethod
    def get_internal_id(cls, item: ThreeDModelClassicResponse | dict) -> int:
        if isinstance(item, dict):
            return item["id"]
        if not item.id:
            raise KeyError("3DModel must have id")
        return item.id

    @classmethod
    def dump_id(cls, id: NameId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def as_str(cls, id: NameId) -> str:
        return sanitize_filename(id.name)

    @classmethod
    def get_required_capability(
        cls, items: Sequence[ThreeDModelClassicRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []
        scope: capabilities.ThreeDAcl.Scope.All | capabilities.ThreeDAcl.Scope.DataSet = (  # type: ignore[valid-type]
            capabilities.ThreeDAcl.Scope.All()
        )
        if items:
            if data_set_ids := {item.data_set_id for item in items or [] if item.data_set_id}:
                scope = capabilities.ThreeDAcl.Scope.DataSet(list(data_set_ids))

        actions = (
            [capabilities.ThreeDAcl.Action.Read]
            if read_only
            else [
                capabilities.ThreeDAcl.Action.Read,
                capabilities.ThreeDAcl.Action.Create,
                capabilities.ThreeDAcl.Action.Update,
                capabilities.ThreeDAcl.Action.Delete,
            ]
        )

        return capabilities.ThreeDAcl(actions, scope)

    def create(self, items: Sequence[ThreeDModelClassicRequest]) -> list[ThreeDModelClassicResponse]:
        return self.client.tool.three_d.models_classic.create(items)

    def retrieve(self, ids: Sequence[NameId]) -> list[ThreeDModelClassicResponse]:
        selected_names = {id_.name for id_ in ids}
        output: list[ThreeDModelClassicResponse] = []
        for models in self.client.tool.three_d.models_classic.iterate(limit=None):
            for model in models:
                if model.name in selected_names:
                    output.append(model)
                    selected_names.discard(model.name)
            if not selected_names:
                break
        return output

    def update(self, items: Sequence[ThreeDModelClassicRequest]) -> list[ThreeDModelClassicResponse]:
        return self.client.tool.three_d.models_classic.update(items)

    def delete(self, ids: Sequence[NameId]) -> int:
        models = self.retrieve(ids)
        internal_ids = [InternalId(id=model.id) for model in models]
        self.client.tool.three_d.models_classic.delete(internal_ids)
        return len(models)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[ThreeDModelClassicResponse]:
        # DataSet filtering is not supported by the API, so we filter client-side.
        data_set_id = self.client.lookup.data_sets.id(data_set_external_id) if data_set_external_id else None
        for models in self.client.tool.three_d.models_classic.iterate(limit=None):
            if data_set_id is None:
                yield from models
                continue
            for model in models:
                if model.data_set_id == data_set_id:
                    yield model

    def drop_data(self, ids: Sequence[NameId]) -> int:
        models = self.retrieve(ids)
        count = 0
        for model in models:
            for revisions in self.client.tool.three_d.revisions_classic.iterate(model_id=model.id, limit=None):
                revision_ids = [ThreeDModelRevisionId(id=r.id, model_id=r.model_id) for r in revisions]
                self.client.tool.three_d.revisions_classic.delete(ids=revision_ids)
                count += len(revisions)
        return count

    def count(self, ids: Sequence[NameId]) -> int:
        models = self.retrieve(ids)
        count = 0
        for model in models:
            for revisions in self.client.tool.three_d.revisions_classic.iterate(model_id=model.id, limit=None):
                count += len(revisions)
        return count

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        """Returns all items that this item requires.

        For example, a TimeSeries requires a DataSet, so this method would return the
        DatasetLoader and identifier of that dataset.
        """
        if "dataSetExternalId" in item:
            yield DataSetsCRUD, ExternalId(external_id=item["dataSetExternalId"])

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> ThreeDModelClassicRequest:
        if ds_external_id := resource.pop("dataSetExternalId", None):
            resource["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        return ThreeDModelClassicRequest.model_validate(resource)

    def dump_resource(
        self, resource: ThreeDModelClassicResponse, local: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump()
        local = local or {}
        if data_set_id := dumped.pop("dataSetId", None):
            dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
        if not dumped.get("metadata") and "metadata" in local:
            # Remove empty metadata {}.
            dumped.pop("metadata", None)
        return dumped



import gzip
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Literal

import httpx
from cognite.client import global_config
from pydantic import BaseModel, Field, JsonValue, TypeAdapter, model_validator

from cognite_toolkit._cdf_tk.client.http_client._exception import ToolkitAPIError
from cognite_toolkit._cdf_tk.utils.useful_types import PrimitiveType

if TYPE_CHECKING:
    from cognite_toolkit._cdf_tk.client.http_client._item_classes import ItemsResultMessage


class HTTPResult(BaseModel):
    def get_success_or_raise(self) -> "SuccessResponse":
        """Raises an exception if any response in the list indicates a failure."""
        if isinstance(self, SuccessResponse):
            return self
        elif isinstance(self, FailedResponse):
            raise ToolkitAPIError(
                f"Request failed with status code {self.status_code}: {self.error.message}",
                missing=self.error.missing,  # type: ignore[arg-type]
                duplicated=self.error.duplicated,  # type: ignore[arg-type]
                code=self.error.code,
            )
        elif isinstance(self, FailedRequest):
            raise ToolkitAPIError(f"Request failed with error: {self.error}")
        else:
            raise ToolkitAPIError("Unknown HTTPResult2 type")

    def as_item_response(self, item_id: str) -> "ItemsResultMessage":
        # Avoid circular import
        from cognite_toolkit._cdf_tk.client.http_client._item_classes import (
            ItemsFailedRequest,
            ItemsFailedResponse,
            ItemsSuccessResponse,
        )

        if isinstance(self, SuccessResponse):
            return ItemsSuccessResponse(
                status_code=self.status_code, content=self.content, ids=[item_id], body=self.body
            )
        elif isinstance(self, FailedResponse):
            return ItemsFailedResponse(
                status_code=self.status_code,
                ids=[item_id],
                body=self.body,
                error=ErrorDetails(
                    code=self.error.code,
                    message=self.error.message,
                    missing=self.error.missing,
                    duplicated=self.error.duplicated,
                    is_auto_retryable=self.error.is_auto_retryable,
                ),
            )
        elif isinstance(self, FailedRequest):
            return ItemsFailedRequest(ids=[item_id], error_message=self.error)
        else:
            raise ToolkitAPIError(f"Unknown {type(self).__name__} type")


class FailedRequest(HTTPResult):
    error: str


class SuccessResponse(HTTPResult):
    status_code: int
    body: str
    content: bytes

    @property
    def body_json(self) -> dict[str, Any]:
        """Parse the response body as JSON."""
        return TypeAdapter(dict[str, JsonValue]).validate_json(self.body)


class ErrorDetails(BaseModel):
    """This is the expected structure of error details in the CDF API"""

    code: int
    message: str
    missing: list[JsonValue] | None = None
    duplicated: list[JsonValue] | None = None
    is_auto_retryable: bool | None = None

    @classmethod
    def from_response(cls, response: httpx.Response) -> "ErrorDetails":
        """Populate the error details from a httpx response."""
        try:
            res = TypeAdapter(dict[Literal["error"], ErrorDetails]).validate_json(response.text)
        except ValueError:
            return cls(code=response.status_code, message=response.text)
        return res["error"]


class FailedResponse(HTTPResult):
    status_code: int
    body: str
    error: ErrorDetails


class BaseRequestMessage(BaseModel, ABC):
    endpoint_url: str
    method: Literal["GET", "POST", "PATCH", "DELETE", "PUT"]
    connect_attempt: int = 0
    read_attempt: int = 0
    status_attempt: int = 0
    api_version: str | None = None
    disable_gzip: bool = False
    content_length: int | None = None
    content_type: str = "application/json"
    accept: str = "application/json"
    client_timeout: float | None = None
    retry_status: bool = Field(
        default=True, description="Whether to retry failed responses. This overrides the HTTPClient configuration"
    )

    parameters: dict[str, PrimitiveType] | None = None

    @property
    def total_attempts(self) -> int:
        return self.connect_attempt + self.read_attempt + self.status_attempt

    @property
    @abstractmethod
    def content(self) -> str | bytes | None: ...


class RequestMessage(BaseRequestMessage):
    data_content: bytes | None = None
    body_content: dict[str, JsonValue] | None = None

    @model_validator(mode="before")
    def check_data_or_body(cls, values: dict[str, Any]) -> dict[str, Any]:
        if values.get("data_content") is not None and values.get("body_content") is not None:
            raise ValueError("Only one of data_content or body_content can be set.")
        return values

    @property
    def content(self) -> str | bytes | None:
        data: str | bytes | None = None
        if self.data_content is not None:
            data = self.data_content
            if not global_config.disable_gzip and not self.disable_gzip:
                data = gzip.compress(data)
        elif self.body_content is not None:
            # We serialize using pydantic instead of json.dumps. This is because pydantic is faster
            # and handles more complex types such as datetime, float('nan'), etc.
            data = _BODY_SERIALIZER.dump_json(self.body_content)
            if not global_config.disable_gzip and not self.disable_gzip and isinstance(data, bytes):
                data = gzip.compress(data)
        return data


_BODY_SERIALIZER = TypeAdapter(dict[str, JsonValue])



from typing import Any, Literal

from pydantic import Field, model_validator

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
    UpdatableRequestResource,
)

from .identifiers import ExternalId

STREAMLIT_DIRECTORY = "/streamlit-apps/"


class StreamlitFile(BaseModelObject):
    """Base class for Streamlit app with common fields."""

    external_id: str
    name: str
    creator: str
    entrypoint: str | None = None
    description: str | None = None
    published: bool | None = None
    theme: Literal["Light", "Dark"] | None = None
    thumbnail: str | None = None
    data_set_id: int | None = None
    cognite_toolkit_app_hash: str | None = Field(None, alias="cdf-toolkit-app-hash")


class StreamlitRequest(StreamlitFile, UpdatableRequestResource):
    """Request resource for creating/updating Streamlit apps."""

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)

    def dump(
        self, camel_case: bool = True, exclude_extra: bool = False, context: Literal["api", "toolkit"] = "api"
    ) -> dict[str, Any]:
        if context == "toolkit":
            return super().dump(camel_case=camel_case, exclude_extra=exclude_extra)
        metadata = self._as_metadata()
        # Returning the file metadata structure expected by the API.
        return {
            "externalId" if camel_case else "external_id": self.external_id,
            "name": f"{self.name}-source.json",
            "dataSetId" if camel_case else "data_set_id": self.data_set_id,
            "directory": STREAMLIT_DIRECTORY,
            "metadata": metadata,
        }

    def _as_metadata(self) -> dict[str, str]:
        metadata: dict[str, str] = {
            "creator": self.creator,
            "name": self.name,
        }
        if self.published is not None:
            metadata["published"] = str(self.published).lower()
        if self.theme is not None:
            metadata["theme"] = str(self.theme)
        if self.description:
            metadata["description"] = self.description
        if self.thumbnail:
            metadata["thumbnail"] = self.thumbnail
        if self.cognite_toolkit_app_hash:
            metadata["cdf-toolkit-app-hash"] = self.cognite_toolkit_app_hash
        if self.entrypoint:
            metadata["entrypoint"] = self.entrypoint
        return metadata

    def as_update(self, mode: Literal["patch", "replace"]) -> dict[str, Any]:
        update_data: dict[str, Any] = {
            "metadata": {"set": self._as_metadata()},
            "directory": {"set": STREAMLIT_DIRECTORY},
        }
        if self.data_set_id is not None:
            update_data["dataSetId"] = {"set": self.data_set_id}
        elif mode == "replace":
            update_data["dataSetId"] = {"setNull": True}

        return {
            "externalId": self.external_id,
            "update": update_data,
        }


class StreamlitResponse(StreamlitFile, ResponseResource[StreamlitRequest]):
    """Response resource for Streamlit apps."""

    id: int
    created_time: int
    last_updated_time: int
    uploaded_time: int | None = None
    uploaded: bool
    # This field is required in the upload endpoint response, but not in any other file metadata response
    upload_url: str | None = None

    @classmethod
    def request_cls(cls) -> type[StreamlitRequest]:
        return StreamlitRequest

    @model_validator(mode="before")
    def move_metadata(cls, values: dict[str, Any]) -> dict[str, Any]:
        if "metadata" not in values:
            return values
            # Move metadata fields to top level if they exist
        values_copy = values.copy()
        metadata = values_copy.pop("metadata")
        values_copy.update(metadata)
        return values_copy



from typing import Any, Generic, Literal, TypeVar

from pydantic import JsonValue

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    RequestResource,
    ResponseResource,
)

from .data_modeling import NodeReference
from .identifiers import ExternalId

# Unlike other resources, datapoints subscriptions have a separate update resource. This
# is because datapoint subscriptions need to handle add/removing operations for timeSeriesIds and
# instanceIds.

T_Value = TypeVar("T_Value")


class Set(BaseModelObject, Generic[T_Value]):
    set: T_Value


class SetNull(BaseModelObject):
    set_null: Literal[True] = True


class AddRemove(BaseModelObject, Generic[T_Value]):
    add: T_Value | None = None
    remove: T_Value | None = None


class DataPointSubscriptionUpdate(BaseModelObject):
    time_series_ids: AddRemove[list[str]] | Set[list[str]] | None = None
    instance_ids: AddRemove[list[NodeReference]] | Set[list[NodeReference]] | None = None
    name: Set[str] | SetNull | None = None
    description: Set[str] | SetNull | None = None
    data_set_id: Set[int] | SetNull | None = None
    filter: Set[JsonValue] | None = None


class DatapointSubscriptionUpdateRequest(RequestResource):
    external_id: str
    update: DataPointSubscriptionUpdate

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)

    def has_data(self) -> bool:
        """Check if the update request contains any data to update."""
        return self.update.model_dump(exclude_none=True) != {}


class DatapointSubscription(BaseModelObject):
    external_id: str
    name: str | None = None
    description: str | None = None
    data_set_id: int | None = None
    partition_count: int
    filter: JsonValue | None = None

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)


class DatapointSubscriptionRequest(DatapointSubscription, RequestResource):
    time_series_ids: list[str] | None = None
    instance_ids: list[NodeReference] | None = None

    def as_update(self) -> DatapointSubscriptionUpdateRequest:
        if self.time_series_ids is not None:
            raise NotImplementedError("Bug in Toolkit. Trying to update timeSeriesIds without an update object")
        if self.instance_ids is not None:
            raise NotImplementedError("Bug in Toolkit. Trying to update instanceIds without an update object")
        filter_arg: dict[str, Any] = {}
        if self.filter is not None:
            filter_arg["filter"] = Set(set=self.filter)
        return DatapointSubscriptionUpdateRequest(
            external_id=self.external_id,
            update=DataPointSubscriptionUpdate(
                name=SetNull() if self.name is None else Set(set=self.name),
                description=SetNull() if self.description is None else Set(set=self.description),
                data_set_id=SetNull() if self.data_set_id is None else Set(set=self.data_set_id),
                **filter_arg,
            ),
        )


class DatapointSubscriptionResponse(DatapointSubscription, ResponseResource[DatapointSubscriptionRequest]):
    time_series_count: int | None = None
    created_time: int
    last_updated_time: int

    @classmethod
    def request_cls(cls) -> type[DatapointSubscriptionRequest]:
        """Return the class of the corresponding request resource."""
        return DatapointSubscriptionRequest
